#------------------------------------------------------------------------------
# Copyright (c) 2014-2017, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#------------------------------------------------------------------------------
from atom.api import Typed

from enaml.drag_drop import DropAction, DropEvent
from enaml.layout.geometry import Pos

from .QtCore import Qt
from .QtGui import QDropEvent
from .qt_mime_data import QtMimeData


class QtDropEvent(DropEvent):
    """ A Qt implementation of an Enaml DragEvent.

    """
    #: Internal storage for the QDropEvent.
    _q_event = Typed(QDropEvent)

    def __init__(self, event):
        """ Initialize a QtDropEvent.

        Parameters
        ----------
        event : QDropEvent
            The Qt drop event object.

        """
        self._q_event = event

    def pos(self):
        """ Get the current mouse position of the operation.

        Returns
        -------
        result : Pos
            The mouse position of the operation in widget coordinates.

        """
        qpos = self._q_event.pos()
        return Pos(qpos.x(), qpos.y())

    def mime_data(self):
        """ Get the mime data contained in the drag operation.

        Returns
        -------
        result : QtMimeData
            The mime data contained in the drag operation.

        """
        return QtMimeData(self._q_event.mimeData())

    def drop_action(self):
        """ Get the action to be performed by the drop target.

        Returns
        -------
        result : DropAction
            A drop action enum value.

        """
        return DropAction(int(self._q_event.dropAction()))

    def possible_actions(self):
        """ Get the OR'd combination of possible drop actions.

        Returns
        -------
        result : DropAction.Flags
            The combination of possible drop actions.

        """
        return DropAction.Flags(int(self._q_event.possibleActions()))

    def proposed_action(self):
        """ Get the action proposed to be taken by the drop target.

        Returns
        -------
        result : DropAction
            The proposed action for the drop target.

        """
        return DropAction(int(self._q_event.proposedAction()))

    def accept_proposed_action(self):
        """ Accept the event using the proposed drop action.

        """
        self._q_event.acceptProposedAction()

    def set_drop_action(self, action):
        """ Set the drop action to one of the possible actions.

        Parameters
        ----------
        action : DropAction
            The drop action to be performed by the target.

        """
        self._q_event.setDropAction(Qt.DropAction(action))

    def is_accepted(self):
        """ Test whether the event has been accepted.

        Returns
        -------
        result : bool
            True if the event is accepted, False otherwise.

        """
        return self._q_event.isAccepted()

    def set_accepted(self, accepted):
        """ Set the accepted state of the event.

        Parameters
        ----------
        accepted : bool
            The target accepted state of the event.

        """
        self._q_event.setAccepted(accepted)

    def accept(self):
        """ Accept the drag event action.

        """
        self._q_event.accept()

    def ignore(self):
        """ Ignore the drag event action.

        """
        self._q_event.ignore()



#------------------------------------------------------------------------------
# Copyright (c) 2013-2018, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#------------------------------------------------------------------------------
import ast
from types import FunctionType

from .. import enaml_ast
from .base_lexer import syntax_error
from .base_parser import BaseEnamlParser, FakeToken, ast_for_testlist, Load
from .lexer3 import Python3EnamlLexer


class Python3EnamlParser(BaseEnamlParser):
    """Enaml parser supporting Python syntax.

    Main differences from base parser are :

    - support for annotation in function declaration
    - support for raise from syntax
    - support for yield from syntax

    """
    parser_id = '3'

    lexer = Python3EnamlLexer

    context_allowed = set(list(BaseEnamlParser.context_allowed) +
                          [ast.Starred])

    def set_context(self, node, ctx, p):
        """Properly set the context of ast.Starred.

        """
        super(Python3EnamlParser, self).set_context(node, ctx, p)
        if( isinstance(node, ast.Starred) and
                type(node.value) in self.context_allowed):
            self.set_context(node.value, ctx, p)

    def p_decl_funcdef3(self, p):
        ''' decl_funcdef : NAME NAME parameters RETURNARROW test COLON suite '''
        lineno = p.lineno(1)
        if p[1] != 'func':
            syntax_error('invalid syntax', FakeToken(p.lexer.lexer, lineno))
        funcdef = ast.FunctionDef()
        funcdef.name = p[2]
        funcdef.args = p[3]
        funcdef.returns = p[5]
        funcdef.body = p[7]
        funcdef.decorator_list = []
        funcdef.lineno = lineno
        ast.fix_missing_locations(funcdef)
        self._validate_decl_funcdef(funcdef, p.lexer.lexer)
        decl_funcdef = enaml_ast.FuncDef()
        decl_funcdef.lineno = lineno
        decl_funcdef.funcdef = funcdef
        decl_funcdef.is_override = False
        p[0] = decl_funcdef

    def p_small_stmt1(self, p):
        ''' small_stmt : expr_stmt
                       | del_stmt
                       | pass_stmt
                       | flow_stmt
                       | import_stmt
                       | global_stmt
                       | assert_stmt
                       | nonlocal_stmt'''
        p[0] = p[1]

    def p_nonlocal_stmt1(self, p):
        ''' nonlocal_stmt : NONLOCAL NAME'''
        nonlocal_stmt = ast.Nonlocal()
        nonlocal_stmt.names = [p[2]]
        nonlocal_stmt.lineno = p.lineno(1)
        p[0] = nonlocal_stmt

    def p_nonlocal_stmt2(self, p):
        ''' global_stmt : NONLOCAL NAME globals_list '''
        nonlocal_stmt = ast.NonLocal()
        nonlocal_stmt.names = [p[2]] + p[3]
        nonlocal_stmt.lineno = p.lineno(1)
        p[0] = nonlocal_stmt

    def p_raise_stmt3(self, p):
        ''' raise_stmt : RAISE test FROM test '''
        raise_stmt = ast.Raise()
        raise_stmt.exc = p[1]
        raise_stmt.cause = p[4]
        p[0] = raise_stmt

    def p_test_or_star2(self, p):
        ''' test_or_star : star_expr '''
        p[0] = p[1]

    def p_star_expr(self, p):
        ''' star_expr : STAR expr '''
        # We can assume Load as we will explicitely set Store
        p[0] = ast.Starred(value=p[2], ctx=Load)

    def p_yield_expr3(self, p):
        ''' yield_expr : YIELD FROM testlist '''
        value = ast_for_testlist(p[2])
        p[0] = ast.Yield(value=value, lineno=p.lineno(1))

    def p_funcdef2(self, p):
        ''' funcdef : DEF NAME parameters RETURNARROW test COLON suite '''
        funcdef = ast.FunctionDef()
        funcdef.name = p[2]
        funcdef.args = p[3]
        funcdef.returns = p[5]
        funcdef.body = p[7]
        funcdef.decorator_list = []
        funcdef.lineno = p.lineno(1)
        ast.fix_missing_locations(funcdef)
        p[0] = funcdef

    def p_parameters2(self, p):
        ''' parameters : LPAR typedargslist RPAR '''
        p[0] = p[2]

    def p_classdef3(self, p):
        ''' classdef : CLASS NAME LPAR arglist RPAR COLON suite '''
        classdef = ast.ClassDef(keywords=[])
        classdef.name = p[2]

        args = p[4]  # This is an Arguments instance
        classdef.bases = args.args
        classdef.keywords = args.keywords
        classdef.starargs = args.starargs
        classdef.kwargs = args.kwargs

        classdef.body = p[7]
        classdef.decorator_list = []
        classdef.lineno = p.lineno(1)
        ast.fix_missing_locations(classdef)
        p[0] = classdef

    def p_atom11(self, p):
        ''' atom : ELLIPSIS '''
        p[0] = ast.Ellipsis()

    def p_atom12(self, p):
        ''' atom : atom_bytes_list '''
        s = ast.Bytes(s=p[1])
        p[0] = s

    def p_atom_bytes_list1(self, p):
        ''' atom_bytes_list : BYTES '''
        p[0] = p[1]

    def p_atom_bytes_list2(self, p):
        ''' atom_bytes_list : atom_bytes_list BYTES '''
        p[0] = p[1] + p[2]

    def _make_args(self, args, defaults=[], vararg=None, kwonlyargs=[],
                   kw_defaults=[], kwarg=None):
        """Build an ast node for function arguments.

        """
        # This is valid only for Python 3.3
        va = vararg.annotation if vararg else None
        vararg = vararg.arg if vararg else None
        ka = kwarg.annotation if kwarg else None
        kwarg = kwarg.arg if kwarg else None

        return ast.arguments(args=args, defaults=defaults, vararg=vararg,
                             varargannotation=va, kwonlyargs=kwonlyargs,
                             kw_defaults=kw_defaults, kwarg=kwarg,
                             kwargannotation=ka)

    def p_varargslist24(self, p):
        ''' varargslist : fpdef COMMA STAR fpdef kwargslist_list'''
        # def f(a, *args, e, b=1): pass
        klist_args, kdefaults = p[5]
        p[0] = self._make_args([p[1]], vararg=p[4], kwonlyargs=klist_args,
                               kw_defaults=kdefaults)

    def p_varargslist25(self, p):
        ''' varargslist : fpdef COMMA STAR kwargslist_list'''
        # def f(a, *, e, b=1): pass
        klist_args, kdefaults = p[4]
        p[0] = self._make_args([p[1]], kwonlyargs=klist_args,
                               kw_defaults=kdefaults)

    def p_varargslist26(self, p):
        ''' varargslist : fpdef COMMA STAR fpdef kwargslist_list COMMA DOUBLESTAR fpdef '''
        # def f(a, *args, e, b=1, **kwargs): pass
        klist_args, kdefaults = p[5]
        p[0] = self._make_args([p[1]], vararg=p[4], kwonlyargs=klist_args,
                               kw_defaults=kdefaults, kwarg=p[8])

    def p_varargslist27(self, p):
        ''' varargslist : fpdef COMMA STAR kwargslist_list COMMA DOUBLESTAR fpdef '''
        # def f(a, *, e, b=1, **kwargs): pass
        klist_args, kdefaults = p[4]
        p[0] = self._make_args([p[1]], kwonlyargs=klist_args,
                               kw_defaults=kdefaults,  kwarg=p[7])

    def p_varargslist28(self, p):
        ''' varargslist : fpdef varargslist_list COMMA STAR fpdef kwargslist_list '''
        # def f(a, c, d=4, *args, e, b=1): pass
        list_args, defaults = p[2]
        args = [p[1]] + list_args
        klist_args, kdefaults = p[6]
        p[0] = self._make_args(args=args, defaults=defaults, vararg=p[5],
                               kwonlyargs=klist_args, kw_defaults=kdefaults)

    def p_varargslist29(self, p):
        ''' varargslist : fpdef varargslist_list COMMA STAR kwargslist_list '''
        # def f(a, c, d=4, *args, e, b=1): pass
        list_args, defaults = p[2]
        args = [p[1]] + list_args
        klist_args, kdefaults = p[5]
        p[0] = self._make_args(args=args, defaults=defaults,
                               kwonlyargs=klist_args, kw_defaults=kdefaults)

    def p_varargslist30(self, p):
        ''' varargslist : fpdef varargslist_list COMMA STAR fpdef kwargslist_list COMMA DOUBLESTAR fpdef '''
        # def f(a, c, *args, e, b=1, **kwargs): pass
        list_args, defaults = p[2]
        args = [p[1]] + list_args
        klist_args, kdefaults = p[6]
        p[0] = self._make_args(args=args, defaults=defaults, vararg=p[5],
                               kwonlyargs=klist_args, kw_defaults=kdefaults,
                               kwarg=p[9])

    def p_varargslist31(self, p):
        ''' varargslist : fpdef varargslist_list COMMA STAR kwargslist_list COMMA DOUBLESTAR fpdef '''
        # def f(a, c, *, e, b=1, **kwargs): pass
        list_args, defaults = p[2]
        args = [p[1]] + list_args
        klist_args, kdefaults = p[5]
        p[0] = self._make_args(args=args, defaults=defaults,
                               kwonlyargs=klist_args, kw_defaults=kdefaults,
                               kwarg=p[8])

    def p_varargslist32(self, p):
        ''' varargslist : fpdef EQUAL test COMMA STAR fpdef kwargslist_list'''
        # def f(a=1, *args, e, b=1): pass
        klist_args, kdefaults = p[7]
        p[0] = self._make_args([p[1]], defaults=[p[3]], vararg=p[6],
                               kwonlyargs=klist_args, kw_defaults=kdefaults)

    def p_varargslist33(self, p):
        ''' varargslist : fpdef EQUAL test COMMA STAR kwargslist_list'''
        # def f(a=1, *args, e, b=1): pass
        klist_args, kdefaults = p[6]
        p[0] = self._make_args([p[1]], defaults=[p[3]],
                               kwonlyargs=klist_args, kw_defaults=kdefaults)

    def p_varargslist34(self, p):
        ''' varargslist : fpdef EQUAL test COMMA STAR fpdef kwargslist_list COMMA DOUBLESTAR fpdef '''
        # def f(a=1, *args, e, b=1, **kwargs): pass
        klist_args, kdefaults = p[7]
        p[0] = self._make_args([p[1]], defaults=[p[3]], vararg=p[6],
                               kwonlyargs=klist_args, kw_defaults=kdefaults,
                               kwarg=p[10])

    def p_varargslist35(self, p):
        ''' varargslist : fpdef EQUAL test COMMA STAR kwargslist_list COMMA DOUBLESTAR fpdef '''
        # def f(a=1, *, e, b=1, **kwargs): pass
        klist_args, kdefaults = p[6]
        p[0] = self._make_args([p[1]], defaults=[p[3]],
                               kwonlyargs=klist_args, kw_defaults=kdefaults,
                               kwarg=p[9])

    def p_varargslist36(self, p):
        ''' varargslist : fpdef EQUAL test varargslist_list COMMA STAR fpdef kwargslist_list'''
        # def f(a=1, b=2, *args, e, b=1): pass
        list_args, list_defaults = p[4]
        if len(list_args) != len(list_defaults):
            msg = 'non-default argument follows default argument.'
            tok = FakeToken(p.lexer.lexer, p.lineno(2))
            syntax_error(msg, tok)
        klist_args, kdefaults = p[8]
        args = [p[1]] + list_args
        defaults = [p[3]] + list_defaults
        p[0] = self._make_args(args, defaults=defaults, vararg=p[7],
                               kwonlyargs=klist_args, kw_defaults=kdefaults)

    def p_varargslist37(self, p):
        ''' varargslist : fpdef EQUAL test varargslist_list COMMA STAR kwargslist_list'''
        # def f(a=1, b=2, *, e, b=1): pass
        list_args, list_defaults = p[4]
        if len(list_args) != len(list_defaults):
            msg = 'non-default argument follows default argument.'
            tok = FakeToken(p.lexer.lexer, p.lineno(2))
            syntax_error(msg, tok)
        klist_args, kdefaults = p[7]
        args = [p[1]] + list_args
        defaults = [p[3]] + list_defaults
        p[0] = self._make_args(args, defaults=defaults,
                               kwonlyargs=klist_args, kw_defaults=kdefaults)

    def p_varargslist38(self, p):
        ''' varargslist : fpdef EQUAL test varargslist_list COMMA STAR fpdef kwargslist_list COMMA DOUBLESTAR fpdef '''
        # def f(a=1, b=2, *args, e, b=1, **kwargs)
        list_args, list_defaults = p[4]
        if len(list_args) != len(list_defaults):
            msg = 'non-default argument follows default argument.'
            tok = FakeToken(p.lexer.lexer, p.lineno(2))
            syntax_error(msg, tok)
        klist_args, kdefaults = p[8]
        args = [p[1]] + list_args
        defaults = [p[3]] + list_defaults
        p[0] = self._make_args(args, defaults=defaults, vararg=p[7],
                               kwonlyargs=klist_args, kw_defaults=kdefaults,
                               kwarg=p[11])

    def p_varargslist39(self, p):
        ''' varargslist : fpdef EQUAL test varargslist_list COMMA STAR kwargslist_list COMMA DOUBLESTAR fpdef '''
        # def f(a=1, b=2, *, e, b=1, **kwargs)
        list_args, list_defaults = p[4]
        if len(list_args) != len(list_defaults):
            msg = 'non-default argument follows default argument.'
            tok = FakeToken(p.lexer.lexer, p.lineno(2))
            syntax_error(msg, tok)
        klist_args, kdefaults = p[7]
        args = [p[1]] + list_args
        defaults = [p[3]] + list_defaults
        p[0] = self._make_args(args, defaults=defaults,
                               kwonlyargs=klist_args, kw_defaults=kdefaults,
                               kwarg=p[10])

    def p_varargslist40(self, p):
        ''' varargslist : STAR fpdef kwargslist_list '''
        # def f(*args, e, b=1): pass
        klist_args, kdefaults = p[3]
        p[0] = self._make_args([], vararg=p[2],
                               kwonlyargs=klist_args, kw_defaults=kdefaults)

    def p_varargslist41(self, p):
        ''' varargslist : STAR kwargslist_list '''
        # def f(*, e, b=1): pass
        klist_args, kdefaults = p[2]
        p[0] = self._make_args([], kwonlyargs=klist_args,
                               kw_defaults=kdefaults)

    def p_varargslist42(self, p):
        ''' varargslist : STAR fpdef kwargslist_list COMMA DOUBLESTAR fpdef '''
        # def f(*args, e, b=1, **kwargs): pass
        klist_args, kdefaults = p[3]
        p[0] = self._make_args(args=[], vararg=p[2], kwarg=p[6],
                               kwonlyargs=klist_args, kw_defaults=kdefaults)

    def p_varargslist43(self, p):
        ''' varargslist : STAR kwargslist_list COMMA DOUBLESTAR fpdef '''
        # def f(*, e, b=1,  **kwargs): pass
        klist_args, kdefaults = p[2]
        p[0] = self._make_args(args=[], kwarg=p[5],
                               kwonlyargs=klist_args, kw_defaults=kdefaults)

    # The kwargslist_list handlers return a 2-tuple of (args, defaults) lists
    def p_kwargslist_list1(self, p):
        ''' kwargslist_list : COMMA fpdef '''
        p[0] = ([p[2]], [None])

    def p_kwargslist_list2(self, p):
        ''' kwargslist_list : COMMA fpdef EQUAL test '''
        p[0] = ([p[2]], [p[4]])

    def p_kwargslist_list3(self, p):
        ''' kwargslist_list : kwargslist_list COMMA fpdef '''
        list_args, list_defaults = p[1]
        args = list_args + [p[3]]
        defaults = list_defaults + [None]
        p[0] = (args, defaults)

    def p_kwargslist_list4(self, p):
        ''' kwargslist_list : kwargslist_list COMMA fpdef EQUAL test '''
        list_args, list_defaults = p[1]
        args = list_args + [p[3]]
        defaults = list_defaults + [p[5]]
        p[0] = (args, defaults)

    def p_fpdef(self, p):
        ''' fpdef : NAME '''
        p[0] = ast.arg(arg=p[1], annotation=None)

    def p_tfpdef1(self, p):
        ''' tfpdef : NAME '''
        p[0] = ast.arg(arg=p[1], annotation=None)

    def p_tfpdef2(self, p):
        ''' tfpdef : NAME COLON test'''
        p[0] = ast.arg(arg=p[1], annotation=p[3])


def _make_typedarg_rule(f, f_name):
    """Copy a rule and allow for annotations.

    """
    rule = FunctionType(f.__code__, f.__globals__, f_name,
                        f.__defaults__, f.__closure__)

    new_doc = f.__doc__.replace('fpdef', 'tfpdef')
    new_doc = new_doc.replace('varargslist', 'typedargslist')
    rule.__doc__ = new_doc.replace('kwargslist', 'typedkwargslist')
    return rule


for f in (x for x in dir(Python3EnamlParser)
          if ('varargslist' in x) | ('kwargslist' in x)):
    if 'varargslist' in f:
        name = f.replace('varargslist', 'typedargslist')
    else:
        name = f.replace('kwargslist', 'typedkwargslist')
    setattr(Python3EnamlParser, name,
            _make_typedarg_rule(getattr(Python3EnamlParser, f), name))



#------------------------------------------------------------------------------
# Copyright (c) 2013, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#------------------------------------------------------------------------------
from abc import ABCMeta, abstractmethod

from future.utils import with_metaclass

from .QtCore import Qt, QSize, QRect
from .QtWidgets import QLayout, QWidgetItem


class AbstractFlowWidget(with_metaclass(ABCMeta, object)):
    """ An abstract base class which defines the interface for widgets
    which can be used in a QFlowLayout.

    Users of QFlowLayout must register their custom QWidget classes with
    this class in order to use the QFlowLayout.

    """

    @abstractmethod
    def layoutData(self):
        """ An abstractmethod which must be implemented by subclasses.

        Returns
        -------
        result : FlowLayoutData
            The FlowLayoutData instance to use for this widget. The
            same data layout data instance should be returned for
            each call to this method.

        """
        raise NotImplementedError


class FlowLayoutData(object):
    """ The layout data object to use with AbstractFlowWidget instances.

    For performance reasons, there are no runtime checks on the limits
    of the values assigned to this class. Users should ensure that the
    values assigned conform to the documented limits. Users must set
    the `dirty` flag to True before calling `updateGeometry` in order
    for changes to have effect.

    """
    #: Whether or not the computed info for the layout item is dirty.
    #: This must be set to True before calling `updateGeometry` on
    #: the owner widget.
    dirty = True

    #: The flow stretch factor of the layout item. This value controls
    #: the amount of space that is taken up by an expandable item in the
    #: direction of the layout flow, relative to the other items in the
    #: line. The minimum is 0 which means the item should not expand.
    #: There is no maximum. The default is 0.
    stretch = 0

    #: The ortho stretch factor of the layout item. This value controls
    #: the amount of space that is taken up by an expandable item in the
    #: direction orthogonal to the layout flow, relative to other items
    #: in the line. The minimum is 0 which means the item should not
    #: expand. There is no maximum. The default is 0.
    ortho_stretch = 0

    #: The alignment of the layout item in the direction orthogonal to
    #: the layout flow. This must be one of the enums Qt.AlignLeading,
    #: Qt.AlignTrailing, or Qt.AlignCenter.
    alignment = Qt.AlignLeading

    #: The preferred size for the layout item. This size will be used
    #: as the size of the layout item to the extent possible. If this
    #: size is invalid in a particular dimension, the sizeHint of the
    #: item in that direction will be used.
    preferred_size = QSize()

    def __init__(self):
        """ Initialize a FlowLayoutData.

        """
        self.preferred_size = QSize()


class QFlowWidgetItem(QWidgetItem):
    """ A custom QWidgetItem for use with the QFlowLayout.

    """
    #: The FlowLayoutData associated with this widget item. It is a
    #: publically accesible attribute for performance reasons.
    data = None

    def __init__(self, widget, data):
        """ Initialize a QFlowWidgetItem.

        Parameters
        ----------
        widget : QWidget
            The widget to manage with this item.

        data : FlowLayoutData
            The layout data struct associated with this item.

        """
        super(QFlowWidgetItem, self).__init__(widget)
        self.data = data
        self._cached_hint = QSize()
        self._cached_max = QSize()
        self._cached_min = QSize()

    def maximumSize(self):
        """ Reimplemented maximum size computation.

        The max size for a flow widget item is cached and recomputed
        only when the widget item is invalidated.

        """
        if not self._cached_max.isValid():
            self._cached_max = super(QFlowWidgetItem, self).maximumSize()
        return self._cached_max

    def minimumSize(self):
        """ Reimplemented minimum size computation.

        The min size for a flow widget item is cached and recomputed
        only when the widget item is invalidated.

        """
        if not self._cached_min.isValid():
            self._cached_min = super(QFlowWidgetItem, self).minimumSize()
        return self._cached_min

    def sizeHint(self):
        """ Reimplemented size hint computation.

        The size hint for a flow widget item is cached and recomputed
        only when the widget item is invalidated. The size hint is the
        valid union of the preferred size, as indicated by the layout
        data, and the size hint of the widget.

        """
        if not self._cached_hint.isValid():
            hint = super(QFlowWidgetItem, self).sizeHint()
            pref = self.data.preferred_size
            smin = self.minimumSize()
            smax = self.maximumSize()
            if pref.width() != -1:
                pw = max(smin.width(), min(pref.width(), smax.width()))
                hint.setWidth(pw)
            if pref.height() != -1:
                ph = max(smin.height(), min(pref.height(), smax.height()))
                hint.setHeight(ph)
            self._cached_hint = hint
        return self._cached_hint

    def setGeometry(self, rect):
        """ Set the rectangle covered by this layout item.

        This reimplemented method ensures that layout always occurs at
        the origin of the given rect. The default QWidgetItem behavior
        is to center the item in the given space.

        Parameters
        ----------
        rect : QRect
            The rectangle that this layout item should cover.

        """
        if self.isEmpty():
            return
        s = rect.size().boundedTo(self.maximumSize())
        self.widget().setGeometry(rect.x(), rect.y(), s.width(), s.height())

    def invalidate(self):
        """ Invalidate the internal cached data for this widget item.

        The invalidation will only have an effect if the layout data
        associate with this item is marked as dirty.

        """
        if self.data.dirty:
            self._cached_hint = QSize()
            self._cached_min = QSize()
            self.data.dirty = False


class _LayoutRow(object):
    """ A private class used by QFlowLayout.

    This class accumulates information about a row of items as the items
    are added to the row. Instances of the this class are created by the
    QFlowLayout during a layout pass. For performance reasons, there are
    several publically accesible attributes. See their documentation for
    restrictions on their use.

    """
    #: The height to use for laying out the row. This attribute is
    #: modified directly by the layout as it distributes the vertical
    #: space amongst the rows.
    layout_height = 0

    #: The minimum height required for the row. This is automatically
    #: updated as items are added to the row. It should be considered
    #: read-only to external users.
    min_height = 0

    #: The minimum width required for the row. This is automatically
    #: updated as items are added to the row. It should be considered
    #: read-only to external users.
    min_width = 0

    #: The desired height for the row. This is automatically updated as
    #: items are added to the row. It should be considered read-only to
    #: external users.
    hint_height = 0

    #: The desired width for the row. This is automatically updated as
    #: items are added to the row. It should be considered read-only to
    #: external users.
    hint_width = 0

    #: The vertical stretch factor for the row. This is automatically
    #: updated as items are added to the row. It should be considered
    #: read-only by external users.
    stretch = 0

    def __init__(self, width, options):
        """ Initialize a layout row.

        Parameters
        ----------
        width : int
            The width of the layout area.

        options : _LayoutOptions
            The options in effect for the layout.

        """
        self._width = width
        self._options = options
        self._items_stretch = 0
        self._items = []

    @property
    def diff_height(self):
        """ A read-only property which computes the difference between
        the desired height and the minimum height.

        """
        return self.hint_height - self.min_height

    def add_item(self, item):
        """ Add an item to the layout row.

        Parameters
        ----------
        item : QFlowWidgetItem
            The flow widget item to add to the layout row.

        Returns
        -------
        result : bool
            True if the items was added. False if the item was not
            added. An item will not be added if the row already
            contains an item and adding another item would cause
            it to overflow.

        """
        min_size = item.minimumSize()
        hint_size = item.sizeHint()
        n = len(self._items)
        s = self._options.h_spacing
        if n > 0 and (self.hint_width + s + hint_size.width()) > self._width:
            return False
        self.min_height = max(self.min_height, min_size.height())
        self.hint_height = max(self.hint_height, hint_size.height())
        self.stretch = max(self.stretch, item.data.ortho_stretch)
        self.min_width += min_size.width()
        self.hint_width += hint_size.width()
        self._items_stretch += item.data.stretch
        if n > 0:
            self.min_width += s
            self.hint_width += s
        self._items.append(item)
        return True

    def layout(self, x, y):
        """ Layout the row using the given starting coordinates.

        Parameters
        ----------
        x : int
            The x coordinate of the row origin.

        y : int
            The y coordinate of the row origin.

        """
        opts = self._options
        layout_width = self._width
        layout_height = self.layout_height
        delta = self._width - self.hint_width
        items = self._items

        # Short circuit the case where there is negative extra space.
        # This means that there must be only a single item in the row,
        # in which case the width may shrink to the minimum if needed.
        if delta < 0:
            assert len(items) == 1
            item = items[0]
            w = max(layout_width, item.minimumSize().width())
            if item.data.ortho_stretch > 0:
                h = min(layout_height, item.maximumSize().height())
            else:
                h = min(layout_height, item.sizeHint().height())
            delta_h = layout_height - h
            if delta_h > 0:
                align = item.data.alignment
                if align == QFlowLayout.AlignTrailing:
                    y += delta_h
                elif align == QFlowLayout.AlignCenter:
                    y += delta_h / 2
            item.setGeometry(QRect(x, y, w, h))
            return

        # Reversing the items reverses the layout direction. All of the
        # computation up to this point has be independent of direction.
        if opts.direction == QFlowLayout.RightToLeft:
            items.reverse()

        # Precompute a map of starting widths for the items. These will
        # be progressively modified as the delta space is distributed.
        widths = {}
        for item in items:
            widths[item] = item.sizeHint().width()

        # If the flow stretch for the row is greater than zero. Then
        # there exists an item or items which have flow stretch. It's
        # not sufficient to simply distribute the delta space according
        # to relative stretch factors, because an item may have a max
        # width which is less than the adjusted width. This causes the
        # rest of the adjustments to be invalid, yielding a potential
        # O(n^2) solution. Instead, the items which can stretch are
        # sorted according to the differences between their desired
        # width and max width. When distributing the delta space in this
        # order, any unused space from an item is added back to the pool
        # and its stretch factor removed from further computation. This
        # gives an O(n log n) solution to the problem. This algorithm
        # iteratively removes space from the delta, so that the alignment
        # pass below operates on the adjusted free space amount.
        items_stretch = self._items_stretch
        if items_stretch > 0:
            diffs = []
            for item in items:
                if item.data.stretch > 0:
                    h = item.sizeHint().width()
                    m = item.maximumSize().width()
                    diffs.append((m - h, item))
            diffs.sort()
            for ignored, item in diffs:
                item_stretch = item.data.stretch
                max_width = item.maximumSize().width()
                d = item_stretch * delta / items_stretch
                items_stretch -= item_stretch
                item_width = widths[item]
                if item_width + d > max_width:
                    widths[item] = max_width
                    delta -= max_width - item_width
                else:
                    widths[item] = item_width + d
                    delta -= d

        # The widths of all items are now computed. Any leftover delta
        # space is used for alignment purposes. This is accomplished by
        # shifting the starting location and, in the case of justify,
        # adding to the horizontal space value.
        start_x = x
        space = opts.h_spacing
        if opts.alignment == QFlowLayout.AlignLeading:
            if opts.direction == QFlowLayout.RightToLeft:
                start_x += delta
        elif opts.alignment == QFlowLayout.AlignTrailing:
            if opts.direction == QFlowLayout.LeftToRight:
                start_x += delta
        elif opts.alignment == QFlowLayout.AlignCenter:
            start_x += delta / 2
        else:
            d = delta / (len(items) + 1)
            space += d
            start_x += d

        # Make a final pass over the items and perform the layout. This
        # pass handles the orthogonal alignment of the item if there is
        # any leftover vertical space for the item.
        curr_x = start_x
        for item in items:
            w = widths[item]
            if item.data.ortho_stretch > 0:
                h = min(layout_height, item.maximumSize().height())
            else:
                h = min(layout_height, item.sizeHint().height())
            delta = layout_height - h
            this_y = y
            if delta > 0:
                align = item.data.alignment
                if align == QFlowLayout.AlignTrailing:
                    this_y = y + delta
                elif align == QFlowLayout.AlignCenter:
                    this_y = y + delta / 2
            item.setGeometry(QRect(curr_x, this_y, w, h))
            curr_x += (w + space)


class _LayoutColumn(object):
    """ A private class used by QFlowLayout.

    This class accumulates information about a column of items as the
    items are added to the column. Instances of the this class are
    created by the QFlowLayout during a layout pass. For performance
    reasons, there are several publically accesible attributes. See
    their documentation for restrictions on their use.

    """
    #: The width to use for laying out the column. This attribute is
    #: modified directly by the layout as it distributes the horizontal
    #: space amongst the columns.
    layout_width = 0

    #: The minimum height required for the column. This is automatically
    #: updated as items are added to the column. It should be considered
    #: read-only to external users.
    min_height = 0

    #: The minimum width required for the column. This is automatically
    #: updated as items are added to the column. It should be considered
    #: read-only to external users.
    min_width = 0

    #: The desired height for the column. This is automatically updated
    #: as items are added to the column. It should be considered
    #: read-only to external users.
    hint_height = 0

    #: The desired width for the column. This is automatically updated
    #: as items are added to the column. It should be considered
    #: read-only to external users.
    hint_width = 0

    #: The vertical stretch factor for the column. This is automatically
    #: updated as items are added to the column. It should be considered
    #: read-only by external users.
    stretch = 0

    def __init__(self, height, options):
        """ Initialize a layout column.

        Parameters
        ----------
        height : int
            The height of the layout area.

        options : _LayoutOptions
            The options in effect for the layout.

        """
        self._height = height
        self._options = options
        self._items_stretch = 0
        self._items = []

    @property
    def diff_width(self):
        """ A read-only property which computes the difference between
        the desired width and the minimum width.

        """
        return self.hint_width - self.min_width

    def add_item(self, item):
        """ Add an item to the layout column.

        Parameters
        ----------
        item : QFlowWidgetItem
            The flow widget item to add to the layout column.

        Returns
        -------
        result : bool
            True if the items was added. False if the item was not
            added. An item will not be added if the column already
            contains an item and adding another item would cause
            it to overflow.

        """
        min_size = item.minimumSize()
        hint_size = item.sizeHint()
        n = len(self._items)
        s = self._options.v_spacing
        if n > 0 and self.hint_height + s + hint_size.height() > self._height:
            return False
        self.min_width = max(self.min_width, min_size.width())
        self.hint_width = max(self.hint_width, hint_size.width())
        self.stretch = max(self.stretch, item.data.ortho_stretch)
        self.min_height += min_size.height()
        self.hint_height += hint_size.height()
        self._items_stretch += item.data.stretch
        if n > 0:
            self.min_height += s
            self.hint_height += s
        self._items.append(item)
        return True

    def layout(self, x, y):
        """ Layout the row using the given starting coordinates.

        Parameters
        ----------
        x : int
            The x coordinate of the column origin.

        y : int
            The y coordinate of the column origin.

        """
        opts = self._options
        layout_height = self._height
        layout_width = self.layout_width
        delta = self._height - self.hint_height
        items = self._items

        # Short circuit the case where there is negative extra space.
        # This means that there must be only a single item in the column,
        # in which case the height may shrink to the minimum if needed.
        if delta < 0:
            assert len(items) == 1
            item = items[0]
            h = max(layout_height, item.minimumSize().height())
            if item.data.ortho_stretch > 0:
                w = min(layout_width, item.maximumSize().width())
            else:
                w = min(layout_width, item.sizeHint().width())
            delta_w = layout_width - w
            if delta_w > 0:
                align = item.data.alignment
                if align == QFlowLayout.AlignTrailing:
                    x += delta_w
                elif align == QFlowLayout.AlignCenter:
                    x += delta_w / 2
            item.setGeometry(QRect(x, y, w, h))
            return

        # Reversing the items reverses the layout direction. All of the
        # computation up to this point has be independent of direction.
        if opts.direction == QFlowLayout.BottomToTop:
            items.reverse()

        # Precompute a map of starting heights for the items. These will
        # be progressively modified as the delta space is distributed.
        heights = {}
        for item in items:
            heights[item] = item.sizeHint().height()

        # See the long comment in _LayoutRow for the explanation about
        # this section of code. This section is simply the transpose.
        items_stretch = self._items_stretch
        if items_stretch > 0:
            diffs = []
            for item in items:
                if item.data.stretch > 0:
                    h = item.sizeHint().height()
                    m = item.maximumSize().height()
                    diffs.append((m - h, item))
            diffs.sort()
            for ignored, item in diffs:
                item_stretch = item.data.stretch
                max_height = item.maximumSize().height()
                d = item_stretch * delta / items_stretch
                items_stretch -= item_stretch
                item_height = heights[item]
                if item_height + d > max_height:
                    heights[item] = max_height
                    delta -= max_height - item_height
                else:
                    heights[item] = item_height + d
                    delta -= d

        # The heights of all items are now computed. Any leftover delta
        # space is used for alignment purposes. This is accomplished by
        # shifting the starting location and, in the case of justify,
        # adding to the vertical space value.
        start_y = y
        space = opts.v_spacing
        if opts.alignment == QFlowLayout.AlignLeading:
            if opts.direction == QFlowLayout.BottomToTop:
                start_y += delta
        elif opts.alignment == QFlowLayout.AlignTrailing:
            if opts.direction == QFlowLayout.TopToBottom:
                start_y += delta
        elif opts.alignment == QFlowLayout.AlignCenter:
            start_y += delta / 2
        else:
            d = delta / (len(items) + 1)
            space += d
            start_y += d

        # Make a final pass over the items and perform the layout. This
        # pass handles the orthogonal alignment of the item if there is
        # any leftover horizontal space for the item.
        curr_y = start_y
        for item in items:
            h = heights[item]
            if item.data.ortho_stretch > 0:
                w = min(layout_width, item.maximumSize().width())
            else:
                w = min(layout_width, item.sizeHint().width())
            delta = layout_width - w
            this_x = x
            if delta > 0:
                align = item.data.alignment
                if align == QFlowLayout.AlignTrailing:
                    this_x = x + delta
                elif align == QFlowLayout.AlignCenter:
                    this_x = x + delta / 2
            item.setGeometry(QRect(this_x, curr_y, w, h))
            curr_y += (h + space)


class QFlowLayout(QLayout):
    """ A custom QLayout which implements a flowing wraparound layout.

    """
    #: Lines are filled from left to right and stacked top to bottom.
    LeftToRight = 0

    #: Lines are filled from right to left and stacked top to bottom.
    RightToLeft = 1

    #: Lines are filled from top to bottom and stacked left to right.
    TopToBottom = 2

    #: Lines are filled from bottom to top and stacked left to right.
    BottomToTop = 3

    #: Lines are aligned to their leading edge.
    AlignLeading = 4

    #: Lines are aligned to their trailing edge.
    AlignTrailing = 5

    #: Lines are aligned centered within any extra space.
    AlignCenter = 6

    #: Lines are aligned justified within any extra space.
    AlignJustify = 7

    def __init__(self):
        """ Initialize a QFlowLayout.

        """
        super(QFlowLayout, self).__init__()
        self._items = []
        self._options = _LayoutOptions()
        self._cached_w = -1
        self._cached_hfw = -1
        self._cached_min = None
        self._cached_hint = None
        self._wfh_size = None

    def addWidget(self, widget):
        """ Add a widget to the end of the flow layout.

        Parameters
        ----------
        widget : AbstractFlowWidget
            The flow widget to add to the layout.

        """
        self.insertWidget(self.count(), widget)

    def insertWidget(self, index, widget):
        """ Insert a widget into the flow layout.

        Parameters
        ----------
        index : int
            The index at which to insert the widget.

        widget : AbstractFlowWidget
            The flow widget to insert into the layout.

        """
        assert isinstance(widget, AbstractFlowWidget), 'invalid widget type'
        self.addChildWidget(widget)
        item = QFlowWidgetItem(widget, widget.layoutData())
        self._items.insert(index, item)
        widget.show()
        self.invalidate()

    def direction(self):
        """ Get the direction of the flow layout.

        Returns
        -------
        result : QFlowLayout direction enum
            The direction of the layout. The default is LeftToRight.

        """
        return self._options.direction

    def setDirection(self, direction):
        """ Set the direction of the flow layout.

        Parameters
        ----------
        direction : QFlowLayout direction enum
            The desired flow direction of the layout.

        """
        self._options.direction = direction
        self.invalidate()

    def alignment(self):
        """ Get the alignment for the lines in the layout.

        Returns
        -------
        result : QFlowLayout alignment enum
            The alignment of the lines in the layout. The default is
            AlignLeading.

        """
        return self._options.alignment

    def setAlignment(self, alignment):
        """ Set the alignment for the lines in the layout.

        Parameters
        ----------
        alignment : QFlowLayout alignment enum
            The desired alignment of the lines in the layout.

        """
        self._options.alignment = alignment
        self.invalidate()

    def horizontalSpacing(self):
        """ Get the horizontal spacing for the layout.

        Returns
        -------
        result : int
            The number of pixels of horizontal space between items in
            the layout. The default is 10px.

        """
        return self._options.h_spacing

    def setHorizontalSpacing(self, spacing):
        """ Set the horizontal spacing for the layout.

        Parameters
        ----------
        spacing : int
            The number of pixels of horizontal space to place between
            items in the layout.

        """
        self._options.h_spacing = spacing
        self.invalidate()

    def verticalSpacing(self):
        """ Get the vertical spacing for the layout.

        Returns
        -------
        result : int
            The number of pixels of vertical space between items in the
            layout. The default is 10px.

        """
        return self._options.v_spacing

    def setVerticalSpacing(self, spacing):
        """ Set the vertical spacing for the layout.

        Parameters
        ----------
        spacing : int
            The number of pixels of vertical space to place between
            items in the layout.

        """
        self._options.v_spacing = spacing
        self.invalidate()

    def hasHeightForWidth(self):
        """ Whether the height of the layout depends on its width.

        Returns
        -------
        result : bool
            True if the flow direction is horizontal, False otherwise.

        """
        return self._options.direction in (self.LeftToRight, self.RightToLeft)

    def heightForWidth(self, width):
        """ Get the height of the layout for the given width.

        This value only applies if `hasHeightForWidth` returns True.

        Parameters
        ----------
        width : int
            The width for which to determine a height.

        """
        if self._cached_w != width:
            left, top, right, bottom = self.getContentsMargins()
            adj_width = width - (left + right)
            height = self._doLayout(QRect(0, 0, adj_width, 0), True)
            self._cached_hfw = height + top + bottom
            self._cached_w = adj_width
        return self._cached_hfw

    def addItem(self, item):
        """ A required virtual method implementation.

        This method should not be used. The methods `addWidget` and
        `insertWidget` should be used instead.

        """
        msg = 'Use `addWidget` and `insertWidget` instead.'
        raise NotImplementedError(msg)

    def invalidate(self):
        """ Invalidate the cached values of the layout.

        """
        self._cached_w = -1
        self._cached_hfw = -1
        self._cached_wfh = -1
        self._cached_min = None
        self._cached_hint = None
        for item in self._items:
            item.invalidate()
        super(QFlowLayout, self).invalidate()

    def count(self):
        """ A virtual method implementation which returns the number of
        items in the layout.

        """
        return len(self._items)

    def itemAt(self, idx):
        """ A virtual method implementation which returns the layout item
        for the given index or None if one does not exist.

        """
        items = self._items
        if idx < len(items):
            return items[idx]

    def takeAt(self, idx):
        """ A virtual method implementation which removes and returns the
        item at the given index or None if one does not exist.

        """
        items = self._items
        if idx < len(items):
            item = items[idx]
            del items[idx]
            item.widget().hide()
            # The creation path of the layout items bypasses the virtual
            # wrapper methods, this means that the ownership of the cpp
            # pointer is never transfered to Qt. If the item is returned
            # here it will be delete by Qt, which doesn't own the pointer.
            # A double free occurs once the Python item falls out of scope.
            # To avoid this, this method always returns None and the item
            # cleanup is performed by Python, which owns the cpp pointer.

    def setGeometry(self, rect):
        """ Sets the geometry of all the items in the layout.

        """
        super(QFlowLayout, self).setGeometry(rect)
        self._doLayout(self.contentsRect())

    def sizeHint(self):
        """ A virtual method implementation which returns the size hint
        for the layout.

        """
        if self._cached_hint is None:
            size = QSize(0, 0)
            for item in self._items:
                size = size.expandedTo(item.sizeHint())
            left, top, right, bottom = self.getContentsMargins()
            size.setWidth(size.width() + left + right)
            size.setHeight(size.height() + top + bottom)
            self._cached_hint = size
        return self._cached_hint

    def minimumSize(self):
        """ A reimplemented method which returns the minimum size hint
        of the layout item widget as the minimum size of the window.

        """
        if self._cached_min is None:
            size = QSize(0, 0)
            for item in self._items:
                size = size.expandedTo(item.minimumSize())
            left, top, right, bottom = self.getContentsMargins()
            size.setWidth(size.width() + left + right)
            size.setHeight(size.height() + top + bottom)
            self._cached_min = size
        # XXX hack! We really need hasWidthForHeight! This doesn't quite
        # work because a QScrollArea internally caches the min size.
        d = self._options.direction
        if d == self.TopToBottom or d == self.BottomToTop:
            m = QSize(self._cached_min)
            if m.width() < self._cached_wfh:
                m.setWidth(self._cached_wfh)
            return m
        return self._cached_min

    #--------------------------------------------------------------------------
    # Private API
    #--------------------------------------------------------------------------
    def _doLayout(self, rect, test=False):
        """ Perform the layout for the given rect.

        Parameters
        ----------
        rect : QRect
            The area to use for performing the layout.

        test : bool, optional
            If True, perform a trial run of the layout without actually
            updating any of the item geometries.

        Returns
        -------
        result : int
            The layout space (width or height) required in the direction
            orthogonal to the layout flow.

        """
        d = self._options.direction
        if d == self.LeftToRight or d == self.RightToLeft:
            res = self._doHorizontalLayout(rect, test)
        else:
            res = self._doVerticalLayout(rect, test)
            # XXX hack! we need hasWidthForHeight
            self._cached_wfh = res + rect.x()
        return res

    def _doHorizontalLayout(self, rect, test):
        """ Perform the layout for a horizontal flow direction.

        The method signature is identical to the `_doLayout` method.

        """
        # Walk over the items and create the layout rows.
        rows = []
        width = rect.width()
        opts = self._options
        for item in self._items:
            if len(rows) == 0:
                row = _LayoutRow(width, opts)
                row.add_item(item)
                rows.append(row)
            else:
                row = rows[-1]
                if not row.add_item(item):
                    row = _LayoutRow(width, opts)
                    row.add_item(item)
                    rows.append(row)

        # After collecting rows all of the rows, compute the metrics. If
        # this is a test run, only the minimum height is required.
        space = opts.v_spacing * (len(rows) - 1)
        if test:
            return sum(row.min_height for row in rows) + space

        min_height = space
        hint_height = space
        total_diff = 0
        stretch = 0
        for row in rows:
            min_height += row.min_height
            hint_height += row.hint_height
            total_diff += row.diff_height
            stretch += row.stretch

        # Make an initial pass to distribute extra space to rows which
        # lie between their minimum height and desired height.
        height = rect.height()
        play_space = max(0, height - min_height)
        diff_space = max(total_diff, 1)  # Guard against divide by zero
        layout_height = 0
        for row in rows:
            d = play_space * row.diff_height / diff_space
            row.layout_height = min(row.min_height + d, row.hint_height)
            layout_height += row.layout_height
        layout_height += space

        # Make a second pass to distribute remaining space to rows
        # which with a stretch factor greater than zero.
        remaining = height - layout_height
        if remaining > 0 and stretch > 0:
            for row in rows:
                if row.stretch > 0:
                    row.layout_height += remaining * row.stretch / stretch

        # Make a final pass to layout the rows, computing the overall
        # final layout height along the way.
        final_height = 0
        x = rect.x()
        curr_y = rect.y()
        v_space = opts.v_spacing
        for row in rows:
            row.layout(x, curr_y)
            d = row.layout_height + v_space
            final_height += d
            curr_y += d

        return final_height

    def _doVerticalLayout(self, rect, test):
        """ Perform the layout for a vertical flow direction.

        The method signature is identical to the `_doLayout` method.

        """
        # Walk over the items and create the layout columns.
        cols = []
        height = rect.height()
        opts = self._options
        for item in self._items:
            if len(cols) == 0:
                col = _LayoutColumn(height, opts)
                col.add_item(item)
                cols.append(col)
            else:
                col = cols[-1]
                if not col.add_item(item):
                    col = _LayoutColumn(height, opts)
                    col.add_item(item)
                    cols.append(col)

        # After collecting rows all of the columns, compute the metrics.
        # If this is a test run, only the minimum width is required.
        space = opts.h_spacing * (len(cols) - 1)
        if test:
            return sum(col.min_width for col in cols) + space

        min_width = space
        hint_width = space
        total_diff = 0
        stretch = 0
        for col in cols:
            min_width += col.min_width
            hint_width += col.hint_width
            total_diff += col.diff_width
            stretch += col.stretch

        # Make an initial pass to distribute extra space to columns
        # which lie between their minimum width and desired width.
        width = rect.width()
        play_space = max(0, width - min_width)
        diff_space = max(total_diff, 1)  # Guard against divide by zero
        layout_width = 0
        for col in cols:
            d = play_space * col.diff_width / diff_space
            col.layout_width = min(col.min_width + d, col.hint_width)
            layout_width += col.layout_width
        layout_width += space

        # Make a second pass to distribute remaining space to columns
        # which with a stretch factor greater than zero.
        remaining = width - layout_width
        if remaining > 0 and stretch > 0:
            for col in cols:
                if col.stretch > 0:
                    col.layout_width += remaining * col.stretch / stretch

        # Make a final pass to layout the columns, computing the overall
        # final layout width along the way.
        final_width = 0
        y = rect.y()
        curr_x = rect.x()
        h_space = opts.h_spacing
        for col in cols:
            col.layout(curr_x, y)
            d = col.layout_width + h_space
            final_width += d
            curr_x += d

        return final_width


class _LayoutOptions(object):
    """ A private class used by QFlowLayout to store layout options.

    """
    #: The flow direction of the layout.
    direction = QFlowLayout.LeftToRight

    #: The alignment of a line in the layout.
    alignment = QFlowLayout.AlignLeading

    #: The horizontal spacing between items or lines in the layout.
    h_spacing = 10

    #: The vertical spacing between items or lines in the layout.
    v_spacing = 10



#------------------------------------------------------------------------------
# Copyright (c) 2013, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#------------------------------------------------------------------------------
from atom.api import (
    Coerced, Event, Unicode, Bool, Range, Typed, ForwardTyped, observe
)

from enaml.application import deferred_call
from enaml.core.declarative import d_
from enaml.icon import Icon
from enaml.layout.geometry import Size

from .container import Container
from .widget import Widget, ProxyWidget


class ProxyDockItem(ProxyWidget):
    """ The abstract definition of a proxy DockItem object.

    """
    #: A reference to the DockItem declaration.
    declaration = ForwardTyped(lambda: DockItem)

    def set_title(self, title):
        raise NotImplementedError

    def set_title_editable(self, editable):
        raise NotImplementedError

    def set_title_bar_visible(self, visible):
        raise NotImplementedError

    def set_icon(self, icon):
        raise NotImplementedError

    def set_icon_size(self, size):
        raise NotImplementedError

    def set_stretch(self, stretch):
        raise NotImplementedError

    def set_closable(self, closable):
        raise NotImplementedError

    def alert(self, level, on, off, repeat, persist):
        raise NotImplementedError


class DockItem(Widget):
    """ A widget which can be docked in a DockArea.

    A DockItem is a widget which can be docked inside of a DockArea. It
    can have at most a single Container child widget.

    """
    #: The title to use in the title bar.
    title = d_(Unicode())

    #: Whether or the not the title is user editable.
    title_editable = d_(Bool(False))

    #: Whether or not the title bar is visible.
    title_bar_visible = d_(Bool(True))

    #: The icon to use in the title bar.
    icon = d_(Typed(Icon))

    #: The size to use for the icon in the title bar.
    icon_size = d_(Coerced(Size, (-1, -1)))

    #: The stretch factor for the item when docked in a splitter.
    stretch = d_(Range(low=0, value=1))

    #: Whether or not the dock item is closable via a close button.
    closable = d_(Bool(True))

    #: An event emitted when the title bar is right clicked.
    title_bar_right_clicked = d_(Event(), writable=False)

    #: An event emitted when the dock item is closed. The item will be
    #: destroyed after this event has completed.
    closed = d_(Event(), writable=False)

    #: A reference to the ProxyDockItem object.
    proxy = Typed(ProxyDockItem)

    def dock_widget(self):
        """ Get the dock widget defined for the dock pane.

        The last child Container is considered the dock widget.

        """
        for child in reversed(self.children):
            if isinstance(child, Container):
                return child

    def alert(self, level, on=250, off=250, repeat=4, persist=False):
        """ Set the alert level on the dock item.

        This will override any currently applied alert level.

        Parameters
        ----------
        level : unicode
            The alert level token to apply to the dock item.

        on : int
            The duration of the 'on' cycle, in ms. A value of -1 means
            always on.

        off : int
            The duration of the 'off' cycle, in ms. If 'on' is -1, this
            value is ignored.

        repeat : int
            The number of times to repeat the on-off cycle. If 'on' is
            -1, this value is ignored.

        persist : bool
            Whether to leave the alert in the 'on' state when the cycles
            finish. If 'on' is -1, this value is ignored.

        """
        if self.proxy_is_active:
            self.proxy.alert(level, on, off, repeat, persist)

    #--------------------------------------------------------------------------
    # Observers
    #--------------------------------------------------------------------------
    @observe('title', 'title_editable', 'title_bar_visible', 'icon',
        'icon_size', 'stretch', 'closable')
    def _update_proxy(self, change):
        """ Update the proxy when the item state changes.

        """
        # The superclass implementation is sufficient.
        super(DockItem, self)._update_proxy(change)

    #--------------------------------------------------------------------------
    # Private API
    #--------------------------------------------------------------------------
    def _item_closed(self):
        """ Called by the proxy when the toolkit item is closed.

        """
        # TODO allow the user to veto the close request
        self.closed()
        deferred_call(self.destroy)



#------------------------------------------------------------------------------
# Copyright (c) 2013, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#------------------------------------------------------------------------------
from types import FunctionType

from atom.api import Atom, Typed

from .dynamicscope import DynamicScope
from .expression_engine import ReadHandler, WriteHandler
from .funchelper import call_func
from .standard_inverter import StandardInverter
from .standard_tracer import StandardTracer


class HandlerMixin(Atom):
    """ A mixin class which provides common handler functionality.

    """
    #: The function to invoke to execute the expression. This value is
    #: provided by the standard operators, and will be appropriate for
    #: the given handler type.
    func = Typed(FunctionType)

    #: The key for the local scope in the storage map. This value is
    #: generated by the compiler and provided by the operators.
    scope_key = Typed(object)

    def get_locals(self, owner):
        """ Get a mapping of locals for expression evaluation.

        Parameters
        ----------
        owner : Declarative
            The object on which the handler is executing.

        Returns
        -------
        result : mapping
            A mapping object to use as the local scope.

        """
        return owner._d_storage.get(self.scope_key) or {}


class StandardReadHandler(ReadHandler, HandlerMixin):
    """ An expression read handler for simple read semantics.

    This handler is used in conjunction with the standard '=' operator.

    """
    def __call__(self, owner, name):
        """ Evaluate and return the expression value.

        """
        func = self.func
        f_globals = func.__globals__
        f_builtins = f_globals['__builtins__']
        f_locals = self.get_locals(owner)
        scope = DynamicScope(owner, f_locals, f_globals, f_builtins)
        return call_func(func, (), {}, scope)


class StandardWriteHandler(WriteHandler, HandlerMixin):
    """ An expression write handler for simple write semantics.

    This handler is used in conjuction with the standard '::' operator.

    """
    def __call__(self, owner, name, change):
        """ Write the change to the expression.

        """
        func = self.func
        f_globals = func.__globals__
        f_builtins = f_globals['__builtins__']
        f_locals = self.get_locals(owner)
        scope = DynamicScope(owner, f_locals, f_globals, f_builtins, change)
        call_func(func, (), {}, scope)


class StandardTracedReadHandler(ReadHandler, HandlerMixin):
    """ An expression read handler which traces code execution.

    This handler is used in conjuction with the standard '<<' operator.

    """
    def __call__(self, owner, name):
        """ Evaluate and return the expression value.

        """
        func = self.func
        f_globals = func.__globals__
        f_builtins = f_globals['__builtins__']
        f_locals = self.get_locals(owner)
        tr = StandardTracer(owner, name)
        scope = DynamicScope(owner, f_locals, f_globals, f_builtins, None, tr)
        return call_func(func, (tr,), {}, scope)


class StandardInvertedWriteHandler(WriteHandler, HandlerMixin):
    """ An expression writer which writes an expression value.

    This handler is used in conjuction with the standard '>>' operator.

    """
    def __call__(self, owner, name, change):
        """ Write the change to the expression.

        """
        func = self.func
        f_globals = func.__globals__
        f_builtins = f_globals['__builtins__']
        f_locals = self.get_locals(owner)
        scope = DynamicScope(owner, f_locals, f_globals, f_builtins)
        inverter = StandardInverter(scope)
        call_func(func, (inverter, change['value']), {}, scope)



# byteplay: CPython assembler/disassembler
# Copyright (C) 2006 Noam Raphael | Version: http://code.google.com/p/byteplay
# Rewritten 2009 Demur Rumed | Version: http://github.com/serprex/byteplay
#                            Screwed the style over, modified stack logic to be more flexible, updated to Python 3
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

__version__ = '1.0'
__all__ = [
    'opmap',
    'opname',
    'opcodes',
    'hasflow',
    'stack_effect',
    'cmp_op',
    'hasname',
    'hasjrel',
    'hasjabs',
    'hasjump',
    'haslocal',
    'hascompare',
    'hasfree',
    'hasconst',
    'Opcode',
    'SetLineno',
    'Label',
    'isopcode',
    'Code']


from sys import version_info
if version_info < (3, 6):
    raise NotImplementedError("Currently only Python versions >3.5 are supported!")

import opcode
from dis import findlabels
from types import CodeType
from enum import Enum


class Opcode(int):
    __str__ = __repr__ = lambda s: opname[s]


opmap = {name.replace('+', '_'): Opcode(code) for name, code in opcode.opmap.items()}
opname = {code: name for name, code in opmap.items()}
opcodes = set(opname)
for cmp_op, hasname in opmap.items():
    globals()[cmp_op] = hasname
    __all__.append(cmp_op)
cmp_op = opcode.cmp_op

del __all__

hasarg = {x for x in opcodes if x >= opcode.HAVE_ARGUMENT}
hasconst = {Opcode(x) for x in opcode.hasconst}
hasname = {Opcode(x) for x in opcode.hasname}
hasjrel = {Opcode(x) for x in opcode.hasjrel}
hasjabs = {Opcode(x) for x in opcode.hasjabs}
hasjump = hasjabs | hasjrel
haslocal = {Opcode(x) for x in opcode.haslocal}
hascompare = {Opcode(x) for x in opcode.hascompare}
hasfree = {Opcode(x) for x in opcode.hasfree}
hascode = {MAKE_FUNCTION}

STOP_CODE = -1
import dis

# Fix bug in Python 3.6.0 (fixed in 3.6.1)
if (3, 6, 0) <= version_info < (3, 6, 1):
    def stack_effect(o, arg):
        return (dis.stack_effect(o, arg) if o != CALL_FUNCTION_EX else
                -2 if arg else -1)
else:
    stack_effect = dis.stack_effect

hasflow = hasjump | {
    POP_BLOCK,
    END_FINALLY,
    BREAK_LOOP,
    RETURN_VALUE,
    RAISE_VARARGS,
    STOP_CODE,
    POP_EXCEPT,
    WITH_CLEANUP_START,
    WITH_CLEANUP_FINISH,
    SETUP_ASYNC_WITH}
coroutine_opcodes = {GET_AWAITABLE, GET_AITER, GET_ANEXT, BEFORE_ASYNC_WITH, SETUP_ASYNC_WITH}


class Label:
    pass


class SetLinenoType:
    def __repr__(self):
        return 'SetLineno'
SetLineno = SetLinenoType()


def isopcode(x):
    return x is not SetLineno and not isinstance(x, Label)


# Flags for codeobject.co_flags, taken from Include/code.h, other flags are no longer used
CO_OPTIMIZED          = 0x0001
CO_NEWLOCALS          = 0x0002
CO_VARARGS            = 0x0004
CO_VARKEYWORDS        = 0x0008
CO_NESTED             = 0x0010
CO_GENERATOR          = 0x0020
CO_NOFREE             = 0x0040
CO_COROUTINE          = 0x0080
CO_ITERABLE_COROUTINE = 0x0100
CO_ASYNC_GENERATOR    = 0x0200

CO_FUTURE_BARRY_AS_BDFL = 0x40000

CO_FUTURE_GENERATOR_STOP = 0x80000


class Code(object):
    """An object which holds all the information which a Python code object
    holds, but in an easy-to-play-with representation

    The attributes are:

    Affecting action
    code - list of 2-tuples: the code
    freevars - list of strings: the free vars of the code (those are names
               of variables created in outer functions and used in the function)
    args - list of strings: the arguments of the code
    kwonly - number of keyword only arguments
    varargs - boolean: Does args end with a '*args' argument
    varkwargs - boolean: Does args end with a '**kwargs' argument
    newlocals - boolean: Should a new local namespace be created
                (True in functions, False for module and exec code)

    force_generator - set CO_GENERATOR in co_flags for generator Code objects without generator-specific code
    Python 3.5:
        force_coroutine - set CO_COROUTINE in co_flags for coroutine Code objects (native coroutines) without coroutine-specific code
        force_iterable_coroutine - set CO_ITERABLE_COROUTINE in co_flags for generator-based coroutine Code objects
        future_generator_stop - set CO_FUTURE_GENERATOR_STOP flag (see PEP-479)
    Python 3.6:
        force_async_generator - set CO_ASYNC_GENERATOR in co_flags

    Not affecting action
    name - string: the name of the code (co_name)
    filename - string: the file name of the code (co_filename)
    firstlineno - int: the first line number (co_firstlineno)
    docstring - string or None: the docstring (the first item of co_consts,
                if it's str)

    code is a list of 2-tuples. The first item is an opcode, or SetLineno, or a
    Label instance. The second item is the argument, if applicable, or None"""

    def __init__(self, code, freevars, args, kwonly, varargs, varkwargs, newlocals,
                 name, filename, firstlineno, docstring,
                 force_generator=False,
                 *, force_coroutine=False, force_iterable_coroutine=False,
                 force_async_generator=False, future_generator_stop=False):
        self.code = code
        self.freevars = freevars
        self.args = args
        self.kwonly = kwonly
        self.varargs = varargs
        self.varkwargs = varkwargs
        self.newlocals = newlocals
        self.name = name
        self.filename = filename
        self.firstlineno = firstlineno
        self.docstring = docstring
        self.force_generator = force_generator
        self.force_coroutine = force_coroutine
        self.force_iterable_coroutine = force_iterable_coroutine
        self.force_async_generator = force_async_generator
        self.future_generator_stop = future_generator_stop

    @staticmethod
    def _findlinestarts(code):
        """Find the offsets in a byte code which are start of lines in the source
        Generate pairs offset,lineno as described in Python/compile.c
        This is a modified version of dis.findlinestarts, which allows multiplelinestarts
        with the same line number"""
        lineno = code.co_firstlineno
        addr = 0
        for byte_incr, line_incr in zip(code.co_lnotab[0::2], code.co_lnotab[1::2]):
            if byte_incr:
                yield addr, lineno
                addr += byte_incr
            lineno += line_incr
        yield addr, lineno

    @classmethod
    def from_code(cls, co):
        """Disassemble a Python code object into a Code object"""
        free_cell_isection = set(co.co_cellvars) & set(co.co_freevars)
        if free_cell_isection:
            print(co.co_name + ': has non-empty co.co_cellvars & co.co_freevars', free_cell_isection)
            return None

        co_code = co.co_code
        labels = {addr: Label() for addr in findlabels(co_code)}
        linestarts = dict(cls._findlinestarts(co))
        cellfree = co.co_cellvars + co.co_freevars
        code = []
        extended_arg = 0
        is_generator = False
        is_coroutine = False

        for i in range(0, len(co_code), 2):
            if i in labels:
                code.append((labels[i], None))
            if i in linestarts:
                code.append((SetLineno, linestarts[i]))
            op = Opcode(co_code[i])
            arg = co_code[i+1] | extended_arg
            if op in hascode:
                lastop, lastarg = code[-2]
                if lastop != LOAD_CONST:
                    raise ValueError("%s should be preceded by LOAD_CONST" % op)

                sub_code = Code.from_code(lastarg)
                if sub_code is None:
                    print(co.co_name + ': has unexpected subcode block')
                    return None

                code[-2] = (LOAD_CONST, sub_code)
            if op == opcode.EXTENDED_ARG:
                extended_arg = arg << 8
            else:
                if op not in hasarg:
                    code.append((op, None))
                    continue
                extended_arg = 0
                byteplay_arg = co.co_consts[arg] if op in hasconst else \
                               co.co_names[arg] if op in hasname else \
                               labels[arg] if op in hasjabs else \
                               labels[i + 2 + arg] if op in hasjrel else \
                               co.co_varnames[arg] if op in haslocal else \
                               cmp_op[arg] if op in hascompare else \
                               cellfree[arg] if op in hasfree else \
                               arg
                code.append((op, byteplay_arg))

            if op == YIELD_VALUE or op == YIELD_FROM:
                is_generator = True

            if op in coroutine_opcodes:
                is_coroutine = True

        varargs = not not co.co_flags & CO_VARARGS
        varkwargs = not not co.co_flags & CO_VARKEYWORDS

        force_coroutine = not is_coroutine and (co.co_flags & CO_COROUTINE)
        force_iterable_coroutine = co.co_flags & CO_ITERABLE_COROUTINE
        force_async_generator = co.co_flags & CO_ASYNC_GENERATOR

        is_generator = False if force_async_generator else is_generator
        force_generator = not is_generator and (co.co_flags & CO_GENERATOR)

        assert not (force_coroutine and force_iterable_coroutine)
        assert not (force_coroutine and force_async_generator)
        assert not (force_iterable_coroutine and force_async_generator)
        future_generator_stop = co.co_flags & CO_FUTURE_GENERATOR_STOP

        return cls(code=code,
                   freevars=co.co_freevars,
                   args=co.co_varnames[:co.co_argcount + varargs + varkwargs + co.co_kwonlyargcount],
                   kwonly=co.co_kwonlyargcount,
                   varargs=varargs,
                   varkwargs=varkwargs,
                   newlocals=not not co.co_flags & CO_NEWLOCALS,
                   name=co.co_name,
                   filename=co.co_filename,
                   firstlineno=co.co_firstlineno,
                   docstring=co.co_consts[0] if co.co_consts and isinstance(co.co_consts[0], str) else None,
                   force_generator=force_generator,
                   force_coroutine=force_coroutine,
                   force_iterable_coroutine=force_iterable_coroutine,
                   force_async_generator=force_async_generator,
                   future_generator_stop=future_generator_stop)

    def __eq__(self, other):
        try:
            if (self.freevars != other.freevars or
                    self.args != other.args or
                    self.kwonly != other.kwonly or
                    self.varargs != other.varargs or
                    self.varkwargs != other.varkwargs or
                    self.newlocals != other.newlocals or
                    self.name != other.name or
                    self.filename != other.filename or
                    self.firstlineno != other.firstlineno or
                    self.docstring != other.docstring or
                    self.force_generator != other.force_generator or
                    len(self.code) != len(other.code)):
                return False
            else:
                if (self.force_coroutine != other.force_coroutine or
                        self.force_iterable_coroutine != other.force_iterable_coroutine or
                        self.future_generator_stop != other.future_generator_stop or
                        self.force_async_generator != other.force_async_generator):
                    return False


            # This isn't trivial due to labels
            lmap = {}
            for (op1, arg1), (op2, arg2) in zip(self.code, other.code):
                if isinstance(op1, Label):
                    if lmap.setdefault(arg1, arg2) is not arg2:
                        return False
                else:
                    if op1 != op2:
                        return False
                    if op1 in hasjump:
                        if lmap.setdefault(arg1, arg2) is not arg2:
                            return False
                    elif arg1 != arg2:
                        return False

            return True
        except:
            return False

    def _compute_stacksize(self, logging=False):
        code = self.code
        label_pos = {op[0]: pos for pos, op in enumerate(code) if isinstance(op[0], Label)}
        # sf_targets are the targets of SETUP_FINALLY opcodes. They are recorded
        # because they have special stack behaviour. If an exception was raised
        # in the block pushed by a SETUP_FINALLY opcode, the block is popped
        # and 3 objects are pushed. On return or continue, the block is popped
        # and 2 objects are pushed. If nothing happened, the block is popped by
        # a POP_BLOCK opcode and 1 object is pushed by a (LOAD_CONST, None)
        # operation
        # Our solution is to record the stack state of SETUP_FINALLY targets
        # as having 3 objects pushed, which is the maximum. However, to make
        # stack recording consistent, the get_next_stacks function will always
        # yield the stack state of the target as if 1 object was pushed, but
        # this will be corrected in the actual stack recording
        sf_targets = {label_pos[arg] for op, arg in code
                      if (op == SETUP_FINALLY or op == SETUP_WITH or op == SETUP_ASYNC_WITH)}

        states = [None] * len(code)
        maxsize = 0

        class BlockType(Enum):
            DEFAULT = 0,
            TRY_FINALLY = 1,
            TRY_EXCEPT = 2,
            LOOP_BODY = 3,
            WITH_BLOCK = 4,
            EXCEPTION = 5,
            SILENCED_EXCEPTION_BLOCK = 6,

        class State:

            def __init__(self, pos=0, stack=(0,), block_stack=(BlockType.DEFAULT,), log=[]):
                self._pos = pos
                self._stack = stack
                self._block_stack = block_stack
                self._log = log

            @property
            def pos(self):
                return self._pos

            @property
            def stack(self):
                return self._stack

            @stack.setter
            def stack(self, val):
                self._stack = val

            def newstack(self, n):
                if self._stack[-1] < -n:
                    raise ValueError("Popped a non-existing element at %s %s" %
                                     (self._pos, code[self._pos - 4: self._pos + 3]))
                return self._stack[:-1] + (self._stack[-1] + n,)

            @property
            def block_stack(self):
                return self._block_stack

            @property
            def log(self):
                return self._log

            def newlog(self, msg):
                if not logging:
                    return None

                log_msg = str(self._pos) + ": " + msg
                if self._stack:
                    log_msg += " (on stack: "
                    log_depth = 2
                    log_depth = min(log_depth, len(self._stack))
                    for pos in range(-1, -log_depth, -1):
                        log_msg += str(self._stack[pos]) + ", "
                    log_msg += str(self._stack[-log_depth])
                    log_msg += ")"
                else:
                    log_msg += " (empty stack)"
                return [log_msg] + self._log

        op = [State()]

        while op:
            cur_state = op.pop()
            o = sum(cur_state.stack)
            if o > maxsize:
                maxsize = o

            o, arg = code[cur_state.pos]

            if isinstance(o, Label):
                if cur_state.pos in sf_targets:
                    cur_state.stack = cur_state.newstack(5)
                if states[cur_state.pos] is None:
                    states[cur_state.pos] = cur_state
                elif states[cur_state.pos].stack != cur_state.stack:
                    check_pos = cur_state.pos + 1
                    while code[check_pos][0] not in hasflow:
                        check_pos += 1
                    if code[check_pos][0] not in (RETURN_VALUE, RAISE_VARARGS, STOP_CODE):
                        if cur_state.pos not in sf_targets:
                            raise ValueError("Inconsistent code at %s %s %s\n%s" %
                                             (cur_state.pos, cur_state.stack, states[cur_state.pos].stack,
                                              code[cur_state.pos - 5:cur_state.pos + 4]))
                        else:
                            # SETUP_FINALLY target inconsistent code!
                            #
                            # Since Python 3.2 assigned exception is cleared at the end of
                            # the except clause (named exception handler).
                            # To perform this CPython (checked in version 3.4.3) adds special
                            # bytecode in exception handler which currently breaks 'regularity' of bytecode.
                            # Exception handler is wrapped in try/finally block and POP_EXCEPT opcode
                            # is inserted before END_FINALLY, as a result cleanup-finally block is executed outside
                            # except handler. It's not a bug, as it doesn't cause any problems during execution, but
                            # it breaks 'regularity' and we can't check inconsistency here. Maybe issue should be
                            # posted to Python bug tracker.
                            pass
                    continue
                else:
                    continue

            if o not in (BREAK_LOOP, RETURN_VALUE, RAISE_VARARGS, STOP_CODE):
                next_pos = cur_state.pos + 1

                if not isopcode(o):
                    op += State(next_pos, cur_state.stack, cur_state.block_stack, cur_state.log),

                elif o not in hasflow:
                    if o in hasarg and not isinstance(arg, int):
                        se = stack_effect(o, 0)
                    else:
                        se = stack_effect(o, arg)

                    log = cur_state.newlog("non-flow command (" + str(o) + ", se = " + str(se) + ")")
                    op += State(next_pos, cur_state.newstack(se), cur_state.block_stack, log),

                elif o == FOR_ITER:
                    inside_for_log = cur_state.newlog("FOR_ITER (+1)")
                    op += State(label_pos[arg], cur_state.newstack(-1), cur_state.block_stack, cur_state.log),\
                          State(next_pos, cur_state.newstack(1), cur_state.block_stack, inside_for_log)

                elif o in (JUMP_FORWARD, JUMP_ABSOLUTE):
                    after_jump_log = cur_state.newlog(str(o))
                    op += State(label_pos[arg], cur_state.stack, cur_state.block_stack, after_jump_log),

                elif o in (JUMP_IF_FALSE_OR_POP, JUMP_IF_TRUE_OR_POP):
                    after_jump_log = cur_state.newlog(str(o) + ", jumped")
                    log = cur_state.newlog(str(o) + ", not jumped (-1)")
                    op += State(label_pos[arg], cur_state.stack, cur_state.block_stack, after_jump_log),\
                          State(next_pos, cur_state.newstack(-1), cur_state.block_stack, log)

                elif o in {POP_JUMP_IF_TRUE, POP_JUMP_IF_FALSE}:
                    after_jump_log = cur_state.newlog(str(o) + ", jumped (-1)")
                    log = cur_state.newlog(str(o) + ", not jumped (-1)")
                    op += State(label_pos[arg], cur_state.newstack(-1), cur_state.block_stack, after_jump_log),\
                          State(next_pos, cur_state.newstack(-1), cur_state.block_stack, log)

                elif o == CONTINUE_LOOP:
                    next_stack, next_block_stack = cur_state.stack, cur_state.block_stack
                    last_popped_block = None
                    while next_block_stack[-1] != BlockType.LOOP_BODY:
                        last_popped_block = next_block_stack[-1]
                        next_stack, next_block_stack = next_stack[:-1], next_block_stack[:-1]

                    if next_stack != cur_state.stack:
                        log = cur_state.newlog("CONTINUE_LOOP, from non-loop block")
                    else:
                        log = cur_state.newlog("CONTINUE_LOOP")

                    jump_to_pos = label_pos[arg]
                    if last_popped_block == BlockType.WITH_BLOCK:
                        next_stack = next_stack[:-1] + (next_stack[-1] - 1,)
                    op += State(jump_to_pos, next_stack, next_block_stack, log),

                elif o == SETUP_LOOP:
                    inside_loop_log = cur_state.newlog("SETUP_LOOP (+block)")
                    op += State(label_pos[arg], cur_state.stack, cur_state.block_stack, cur_state.log),\
                          State(next_pos, cur_state.stack + (0,), cur_state.block_stack + (BlockType.LOOP_BODY,), inside_loop_log)

                elif o == SETUP_EXCEPT:
                    inside_except_log = cur_state.newlog("SETUP_EXCEPT, exception (+6, +block)")
                    inside_try_log = cur_state.newlog("SETUP_EXCEPT, try-block (+block)")
                    op += State(label_pos[arg], cur_state.stack + (6,), cur_state.block_stack + (BlockType.EXCEPTION,), inside_except_log),\
                          State(next_pos, cur_state.stack + (0,), cur_state.block_stack + (BlockType.TRY_EXCEPT,), inside_try_log)

                elif o == SETUP_FINALLY:
                    inside_finally_block = cur_state.newlog("SETUP_FINALLY (+1)")
                    inside_try_log = cur_state.newlog("SETUP_FINALLY try-block (+block)")
                    op += State(label_pos[arg], cur_state.newstack(1), cur_state.block_stack, inside_finally_block),\
                          State(next_pos, cur_state.stack + (0,), cur_state.block_stack + (BlockType.TRY_FINALLY,), inside_try_log)

                elif o == POP_BLOCK:
                    log = cur_state.newlog("POP_BLOCK (-block)")
                    op += State(next_pos, cur_state.stack[:-1], cur_state.block_stack[:-1], log),

                elif o == POP_EXCEPT:
                    log = cur_state.newlog("POP_EXCEPT (-block)")
                    op += State(next_pos, cur_state.stack[:-1], cur_state.block_stack[:-1], log),

                elif o == END_FINALLY:
                    if cur_state.block_stack[-1] == BlockType.SILENCED_EXCEPTION_BLOCK:
                        log = cur_state.newlog("END_FINALLY pop silenced exception block (-block)")
                        op += State(next_pos, cur_state.stack[:-1], cur_state.block_stack[:-1], log),
                    elif cur_state.block_stack[-1] == BlockType.EXCEPTION:
                        # Reraise exception
                        pass
                    else:
                        log = cur_state.newlog("END_FINALLY (-6)")
                        op += State(next_pos, cur_state.newstack(-6), cur_state.block_stack, log),

                elif o == SETUP_WITH or o == SETUP_ASYNC_WITH:
                    inside_with_block = cur_state.newlog("SETUP_WITH, with-block (+1, +block)")
                    inside_finally_block = cur_state.newlog("SETUP_WITH, finally (+1)")
                    op += State(label_pos[arg], cur_state.newstack(1), cur_state.block_stack, inside_finally_block),\
                          State(next_pos, cur_state.stack + (1,), cur_state.block_stack + (BlockType.WITH_BLOCK,), inside_with_block)

                elif o == WITH_CLEANUP_START:
                    # There is special case when 'with' __exit__ function returns True,
                    # that's the signal to silence exception, in this case additional element is pushed
                    # and next END_FINALLY command won't reraise exception.
                    # Emulate this situation on WITH_CLEANUP_START with creating special block which will be
                    # handled differently by WITH_CLEANUP_FINISH and will cause END_FINALLY not to reraise exception.
                    log = cur_state.newlog("WITH_CLEANUP_START (+1)")
                    silenced_exception_log = cur_state.newlog("WITH_CLEANUP_START silenced_exception (+block)")
                    op += State(next_pos, cur_state.newstack(1), cur_state.block_stack, log),\
                          State(next_pos, cur_state.newstack(-7) + (9,), cur_state.block_stack + (BlockType.SILENCED_EXCEPTION_BLOCK,), silenced_exception_log)

                elif o == WITH_CLEANUP_FINISH:
                    if cur_state.block_stack[-1] == BlockType.SILENCED_EXCEPTION_BLOCK:
                        # See comment in WITH_CLEANUP_START handler
                        log = cur_state.newlog("WITH_CLEANUP_FINISH silenced_exception (-1)")
                        op += State(next_pos, cur_state.newstack(-1), cur_state.block_stack, log),
                    else:
                        log = cur_state.newlog("WITH_CLEANUP_FINISH (-2)")
                        op += State(next_pos, cur_state.newstack(-2), cur_state.block_stack, log),

                else:
                    raise ValueError("Unhandled opcode %s" % o)

        return maxsize + 6  # for exception raise in deepest place

    def to_code(self, from_function=False):
        """Assemble a Python code object from a Code object"""

        num_fastnames = sum(1 for op, arg in self.code if isopcode(op) and op in haslocal)
        is_function = self.newlocals or num_fastnames > 0 or len(self.args) > 0
        nested = is_function and from_function

        co_flags = {op[0] for op in self.code}

        if not self.force_async_generator:
            is_generator = (self.force_generator or
                            (YIELD_VALUE in co_flags or YIELD_FROM in co_flags)
                            )
        else:
            is_generator = False
        no_free = (not self.freevars) and (not co_flags & hasfree)

        is_native_coroutine = bool(self.force_coroutine or (co_flags & coroutine_opcodes))
        assert not (is_native_coroutine and self.force_iterable_coroutine)
        assert not (is_native_coroutine and self.force_async_generator)

        co_flags =\
            (not(STORE_NAME in co_flags or LOAD_NAME in co_flags or DELETE_NAME in co_flags)) |\
            (self.newlocals and CO_NEWLOCALS) |\
            (self.varargs and CO_VARARGS) |\
            (self.varkwargs and CO_VARKEYWORDS) |\
            (is_generator and CO_GENERATOR) |\
            (no_free and CO_NOFREE) |\
            (nested and CO_NESTED)

        co_flags |= (is_native_coroutine and CO_COROUTINE) |\
                    (self.force_iterable_coroutine and CO_ITERABLE_COROUTINE) |\
                    (self.future_generator_stop and CO_FUTURE_GENERATOR_STOP) |\
                    (self.force_async_generator and CO_ASYNC_GENERATOR)

        co_consts = [self.docstring]
        co_names = []
        co_varnames = list(self.args)
        co_freevars = tuple(self.freevars)

        # Find all cellvars beforehand for two reasons
        # Need the number of them to construct the numeric arg for ops in hasfree
        # Need to put args which are cells in the beginning of co_cellvars
        cellvars = {arg for op, arg in self.code
                    if isopcode(op) and op in hasfree
                    and arg not in co_freevars}
        co_cellvars = [jumps for jumps in self.args if jumps in cellvars]

        def index(seq, item, eq=True, can_append=True):
            for i, x in enumerate(seq):
                if x == item if eq else x is item:
                    return i
            if can_append:
                seq.append(item)
                return len(seq) - 1
            else:
                raise IndexError("Item not found")

        jumps = []
        label_pos = {}
        lastlineno = self.firstlineno
        lastlinepos = 0
        co_code = bytearray()
        co_lnotab = bytearray()

        for i, (op, arg) in enumerate(self.code):
            if isinstance(op, Label):
                label_pos[op] = len(co_code)
            elif op is SetLineno:
                incr_lineno = arg - lastlineno
                incr_pos = len(co_code) - lastlinepos
                lastlineno = arg
                lastlinepos += incr_pos
                if incr_lineno != 0 or incr_pos != 0:
                    while incr_pos > 255:
                        co_lnotab += b"\xFF\0"
                        incr_pos -= 255
                    while incr_lineno > 255:
                        co_lnotab += bytes((incr_pos, 255))
                        incr_pos = 0
                        incr_lineno -= 255
                    if incr_pos or incr_lineno:
                        co_lnotab += bytes((incr_pos, incr_lineno))
            elif op == opcode.EXTENDED_ARG:
                self.code[i + 1][1] |= 1 << 32
            else:
                if op in hasconst:
                    if (isinstance(arg, Code) and
                            i + 2 < len(self.code) and
                            self.code[i + 2][0] in hascode):
                        arg = arg.to_code(from_function=is_function)
                        assert arg is not None
                    arg = index(co_consts, arg, 0)
                elif op in hasname:
                    arg = index(co_names, arg)
                elif op in hasjump:
                    jumps.append((len(co_code), arg))
                    co_code += bytes((0x90, 0, op, 0))
                    continue
                elif op in haslocal:
                    arg = index(co_varnames, arg)
                elif op in hascompare:
                    arg = index(cmp_op, arg, can_append=False)
                elif op in hasfree:
                    try:
                        arg = index(co_freevars, arg, can_append=False) + len(cellvars)
                    except IndexError:
                        arg = index(co_cellvars, arg)
                if arg is None:
                    arg = 0
                if arg > 0xFFFFFF:
                    co_code += (opcode.EXTENDED_ARG | (arg >> 16 & 0xFF00)).to_bytes(2, "little")
                if arg > 0xFFFF:
                    co_code += (opcode.EXTENDED_ARG | (arg >> 8 & 0xFF00)).to_bytes(2, "little")
                if arg > 0xFF:
                    co_code += (opcode.EXTENDED_ARG | (arg & 0xFF00)).to_bytes(2, "little")
                co_code += (op | (arg & 0xFF) << 8).to_bytes(2, "little")

        for pos, label in jumps:
            jump = label_pos[label]
            if co_code[pos+2] in hasjrel:
                jump -= pos + 4
            if jump > 0xFFFF:
                raise NotImplementedError("Multiple EXTENDED_ARG jumps not implemented")
            co_code[pos + 3] = jump & 0xFF
            co_code[pos + 1] = jump >> 8 & 0xFF

        co_argcount = len(self.args) - self.varargs - self.varkwargs - self.kwonly
        co_stacksize = self._compute_stacksize()

        return CodeType(co_argcount, self.kwonly, len(co_varnames), co_stacksize, co_flags,
                        bytes(co_code), tuple(co_consts), tuple(co_names), tuple(co_varnames),
                        self.filename, self.name, self.firstlineno, bytes(co_lnotab), co_freevars,
                        tuple(co_cellvars))



#------------------------------------------------------------------------------
# Copyright (c) 2013-2017, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#------------------------------------------------------------------------------
def action_factory():
    from .qt_action import QtAction
    return QtAction


def action_group_factory():
    from .qt_action_group import QtActionGroup
    return QtActionGroup


def calendar_factory():
    from .qt_calendar import QtCalendar
    return QtCalendar


def check_box_factory():
    from .qt_check_box import QtCheckBox
    return QtCheckBox


def color_dialog_factory():
    from .qt_color_dialog import QtColorDialog
    return QtColorDialog


def combo_box_factory():
    from .qt_combo_box import QtComboBox
    return QtComboBox


def container_factory():
    from .qt_container import QtContainer
    return QtContainer


def control_factory():
    from .qt_control import QtControl
    return QtControl


def date_selector_factory():
    from .qt_date_selector import QtDateSelector
    return QtDateSelector


def datetime_selector_factory():
    from .qt_datetime_selector import QtDatetimeSelector
    return QtDatetimeSelector


def dialog_factory():
    from .qt_dialog import QtDialog
    return QtDialog


def dock_area_factory():
    from .qt_dock_area import QtDockArea
    return QtDockArea


def dock_item_factory():
    from .qt_dock_item import QtDockItem
    return QtDockItem


def dock_pane_factory():
    from .qt_dock_pane import QtDockPane
    return QtDockPane


def dual_slider_factory():
    from .qt_dual_slider import QtDualSlider
    return QtDualSlider


def field_factory():
    from .qt_field import QtField
    return QtField


def file_dialog_factory():
    from .qt_file_dialog import QtFileDialog
    return QtFileDialog


def file_dialog_ex_factory():
    from .qt_file_dialog_ex import QtFileDialogEx
    return QtFileDialogEx


def flow_area_factory():
    from .qt_flow_area import QtFlowArea
    return QtFlowArea


def flow_item_factory():
    from .qt_flow_item import QtFlowItem
    return QtFlowItem


def focus_tracker_factory():
    from .qt_focus_tracker import QtFocusTracker
    return QtFocusTracker


def group_box_factory():
    from .qt_group_box import QtGroupBox
    return QtGroupBox


def html_factory():
    from .qt_html import QtHtml
    return QtHtml


def image_view_factory():
    from .qt_image_view import QtImageView
    return QtImageView


def ipython_console_factory():
    from .qt_ipython_console import QtIPythonConsole
    return QtIPythonConsole


def label_factory():
    from .qt_label import QtLabel
    return QtLabel


def main_window_factory():
    from .qt_main_window import QtMainWindow
    return QtMainWindow


def mdi_area_factory():
    from .qt_mdi_area import QtMdiArea
    return QtMdiArea


def mdi_window_factory():
    from .qt_mdi_window import QtMdiWindow
    return QtMdiWindow


def menu_factory():
    from .qt_menu import QtMenu
    return QtMenu


def menu_bar_factory():
    from .qt_menu_bar import QtMenuBar
    return QtMenuBar


def mpl_canvas_factory():
    from .qt_mpl_canvas import QtMPLCanvas
    return QtMPLCanvas


def multiline_field_factory():
    from .qt_multiline_field import QtMultilineField
    return QtMultilineField


def notebook_factory():
    from .qt_notebook import QtNotebook
    return QtNotebook


def object_combo_factory():
    from .qt_object_combo import QtObjectCombo
    return QtObjectCombo


def page_factory():
    from .qt_page import QtPage
    return QtPage


def popup_view_factory():
    from .qt_popup_view import QtPopupView
    return QtPopupView


def push_button_factory():
    from .qt_push_button import QtPushButton
    return QtPushButton


def progress_bar_factory():
    from .qt_progress_bar import QtProgressBar
    return QtProgressBar


def radio_button_factory():
    from .qt_radio_button import QtRadioButton
    return QtRadioButton


def raw_widget_factory():
    from .qt_raw_widget import QtRawWidget
    return QtRawWidget


def scintilla_factory():
    from .qt_scintilla import QtScintilla
    return QtScintilla


def scroll_area_factory():
    from .qt_scroll_area import QtScrollArea
    return QtScrollArea


def separator_factory():
    from .qt_separator import QtSeparator
    return QtSeparator


def slider_factory():
    from .qt_slider import QtSlider
    return QtSlider


def spin_box_factory():
    from .qt_spin_box import QtSpinBox
    return QtSpinBox


def split_item_factory():
    from .qt_split_item import QtSplitItem
    return QtSplitItem


def splitter_factory():
    from .qt_splitter import QtSplitter
    return QtSplitter


def stack_factory():
    from .qt_stack import QtStack
    return QtStack


def stack_item_factory():
    from .qt_stack_item import QtStackItem
    return QtStackItem


def status_bar_factory():
    from .qt_status_bar import QtStatusBar
    return QtStatusBar


def status_item_factory():
    from .qt_status_item import QtStatusItem
    return QtStatusItem


def time_selector_factory():
    from .qt_time_selector import QtTimeSelector
    return QtTimeSelector


def timer_factory():
    from .qt_timer import QtTimer
    return QtTimer


def tool_bar_factory():
    from .qt_tool_bar import QtToolBar
    return QtToolBar


def tool_button_factory():
    from .qt_tool_button import QtToolButton
    return QtToolButton


def vtk_canvas_factory():
    from .qt_vtk_canvas import QtVTKCanvas
    return QtVTKCanvas


def web_view_factory():
    from .qt_web_view import QtWebView
    return QtWebView


def window_factory():
    from .qt_window import QtWindow
    return QtWindow


QT_FACTORIES = {
    'Action': action_factory,
    'ActionGroup': action_group_factory,
    'Calendar': calendar_factory,
    'CheckBox': check_box_factory,
    'ColorDialog': color_dialog_factory,
    'ComboBox': combo_box_factory,
    'Container': container_factory,
    'DateSelector': date_selector_factory,
    'DatetimeSelector': datetime_selector_factory,
    'Dialog': dialog_factory,
    'DockArea': dock_area_factory,
    'DockItem': dock_item_factory,
    'DockPane': dock_pane_factory,
    'DualSlider': dual_slider_factory,
    'Field': field_factory,
    'FileDialog': file_dialog_factory,
    'FileDialogEx': file_dialog_ex_factory,
    'FlowArea': flow_area_factory,
    'FlowItem': flow_item_factory,
    'FocusTracker': focus_tracker_factory,
    'GroupBox': group_box_factory,
    'Html': html_factory,
    'ImageView': image_view_factory,
    'IPythonConsole': ipython_console_factory,
    'Label': label_factory,
    'MainWindow': main_window_factory,
    'MdiArea': mdi_area_factory,
    'MdiWindow': mdi_window_factory,
    'Menu': menu_factory,
    'MenuBar': menu_bar_factory,
    'MPLCanvas': mpl_canvas_factory,
    'MultilineField': multiline_field_factory,
    'Notebook': notebook_factory,
    'ObjectCombo': object_combo_factory,
    'Page': page_factory,
    'PopupView': popup_view_factory,
    'PushButton': push_button_factory,
    'ProgressBar': progress_bar_factory,
    'RadioButton': radio_button_factory,
    'RawWidget': raw_widget_factory,
    'Scintilla': scintilla_factory,
    'ScrollArea': scroll_area_factory,
    'Separator': separator_factory,
    'Slider': slider_factory,
    'SpinBox': spin_box_factory,
    'SplitItem': split_item_factory,
    'Splitter': splitter_factory,
    'Stack': stack_factory,
    'StackItem': stack_item_factory,
    'StatusBar': status_bar_factory,
    'StatusItem': status_item_factory,
    'TimeSelector': time_selector_factory,
    'Timer': timer_factory,
    'ToolBar': tool_bar_factory,
    'ToolButton': tool_button_factory,
    'VTKCanvas': vtk_canvas_factory,
    'WebView': web_view_factory,
    'Window': window_factory,
}



#------------------------------------------------------------------------------
# Copyright (c) 2013, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#------------------------------------------------------------------------------
from abc import ABCMeta


from future.utils import with_metaclass
import kiwisolver as kiwi


class LinearSymbolic(with_metaclass(ABCMeta, object)):
    """ An abstract base class for testing linear symbolic interfaces.

    """


LinearSymbolic.register(kiwi.Variable)
LinearSymbolic.register(kiwi.Term)
LinearSymbolic.register(kiwi.Expression)



#------------------------------------------------------------------------------
# Copyright (c) 2013, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#------------------------------------------------------------------------------
from future.builtins import str
from atom.api import (
    Enum, Bool, Callable, List, Unicode, Typed, ForwardTyped, Event
)

from enaml.application import deferred_call
from enaml.core.declarative import d_
from .toolkit_object import ToolkitObject, ProxyToolkitObject


class ProxyFileDialog(ProxyToolkitObject):
    """ The abstract definition of a proxy FileDialog object.

    """
    #: A reference to the FileDialog declaration.
    declaration = ForwardTyped(lambda: FileDialog)

    def open(self):
        raise NotImplementedError

    def exec_(self):
        raise NotImplementedError


class FileDialog(ToolkitObject):
    """ A dialog widget that allows the user to open and save files and
    directories.

    """
    #: The title to use for the dialog.
    title = d_(Unicode())

    #: The mode of the dialog.
    mode = d_(Enum('open_file', 'open_files', 'save_file', 'directory'))

    #: The selected path in the dialog. This value will be used to set
    #: the initial working directory and file, as appropriate, when the
    #: dialog is opened. It will aslo be updated when the dialog is
    #: closed and accepted.
    path = d_(Unicode())

    #: The list of selected paths in the dialog. It will be updated
    #: when the dialog is closed and accepted. It is output only and
    #: is only applicable for the `open_files` mode.
    paths = List(Unicode())

    #: The string filters used to restrict the user's selections.
    filters = d_(List(Unicode()))

    #: The selected filter from the list of filters. This value will be
    #: used as the initial working filter when the dialog is opened. It
    #: will also be updated when the dialog is closed and accepted.
    selected_filter = d_(Unicode())

    #: Whether to use a platform native dialog, when available. This
    #: attribute is deprecated and no longer has any effect. Native
    #: dialogs are always used when available in a given toolkit.
    native_dialog = d_(Bool(True))

    #: An enum indicating if the dialog was accepted or rejected by
    #: the user. It will be updated when the dialog is closed. This
    #: value is output only.
    result = Enum('rejected', 'accepted')

    #: An optional callback which will be invoked when the dialog is
    #: closed. This is a convenience to make it easier to handle a
    #: dialog opened in non-blocking mode. The callback must accept
    #: a single argument, which will be the dialog instance.
    callback = d_(Callable())

    #: An event fired if the dialog is accepted. The payload will be
    #: the selected path.
    accepted = d_(Event(str), writable=False)

    #: An event fired when the dialog is rejected. It has no payload.
    rejected = d_(Event(), writable=False)

    #: An event fired when the dialog is closed. It has no payload.
    closed = d_(Event(), writable=False)

    #: Whether to destroy the dialog widget on close. The default is
    #: True since dialogs are typically used in a transitory fashion.
    destroy_on_close = d_(Bool(True))

    #: A reference to the ProxyFileDialog object.
    proxy = Typed(ProxyFileDialog)

    def open(self):
        """ Open the dialog in a non-blocking fashion.

        This method will always return None. The state of the dialog
        will be updated when the dialog is closed by the user.

        """
        if not self.is_initialized:
            self.initialize()
        self.proxy.open()

    def exec_(self):
        """ Open the dialog in a blocking fashion.

        Returns
        -------
        result : unicode
            The path selected by the user, or an empty string if the
            dialog is cancelled.

        """
        if not self.is_initialized:
            self.initialize()
        self.proxy.exec_()
        if self.result == 'accepted':
            return self.path
        return u''

    #--------------------------------------------------------------------------
    # Utility Methods
    #--------------------------------------------------------------------------
    def _handle_close(self, result, paths, selected_filter):
        """ Called by the proxy object when the dialog is closed.

        Parameters
        ----------
        result : string
            The result of the dialog; either 'accepted' or 'rejected'.

        paths : list
            The list of user selected paths. If the result is 'rejected'
            this should be an empty list.

        selected_filter : unicode
            The user selected name filter. If the result is 'rejected'
            this should be an empty string.

        """
        self.result = result
        if result == 'accepted':
            self.paths = paths
            self.path = paths[0] if paths else u''
            self.selected_filter = selected_filter
            self.accepted(self.path)
        else:
            self.rejected()
        if self.callback:
            self.callback(self)
        self.closed()
        if self.destroy_on_close:
            deferred_call(self.destroy)



# lextab2.py. This file automatically created by PLY (version 3.9). Don't edit!
_tabversion   = '3.8'
_lextokens    = set(('LPAR', 'ENDMARKER', 'LESS', 'AMPEREQUAL', 'CIRCUMFLEX', 'WS', 'WITH', 'MINUS', 'NEWLINE', 'EXCEPT', 'PLUS', 'PERCENTEQUAL', 'ELLIPSIS', 'EQEQUAL', 'RIGHTSHIFTEQUAL', 'EXEC', 'AMPER', 'STRING_START_SINGLE', 'SLASH', 'PASS', 'NAME', 'INDENT', 'MINUSEQUAL', 'ENAMLDEF', 'DEDENT', 'STRING_START_TRIPLE', 'STAR', 'NOTEQUAL', 'PRINT', 'DOUBLESTAR', 'DEF', 'CIRCUMFLEXEQUAL', 'COLON', 'DOUBLECOLON', 'STRING_CONTINUE', 'FOR', 'DOUBLESTAREQUAL', 'ELSE', 'TRY', 'AND', 'LBRACE', 'AS', 'OR', 'LEFTSHIFT', 'CONTINUE', 'NOT', 'LAMBDA', 'RAISE', 'GLOBAL', 'WHILE', 'VBAR', 'RETURN', 'DOT', 'LEFTSHIFTEQUAL', 'TILDE', 'RSQB', 'PERCENT', 'DOUBLESLASH', 'RBRACE', 'EQUAL', 'PLUSEQUAL', 'IMPORT', 'LESSEQUAL', 'LSQB', 'RIGHTARROW', 'GREATER', 'VBAREQUAL', 'BREAK', 'TEMPLATE', 'DEL', 'STAREQUAL', 'ELIF', 'CONST', 'SLASHEQUAL', 'NUMBER', 'RPAR', 'ASSERT', 'STRING_END', 'GREATEREQUAL', 'SEMI', 'DOUBLESLASHEQUAL', 'COMMA', 'CLASS', 'RIGHTSHIFT', 'STRING', 'COLONEQUAL', 'IS', 'YIELD', 'ALIAS', 'FINALLY', 'AT', 'PRAGMA', 'IN', 'IF', 'FROM'))
_lexreflags   = 0
_lexliterals  = ''
_lexstateinfo = {'TRIPLEQ2': 'exclusive', 'TRIPLEQ1': 'exclusive', 'INITIAL': 'inclusive', 'SINGLEQ2': 'exclusive', 'SINGLEQ1': 'exclusive'}
_lexstatere   = {'TRIPLEQ2': [('(?P<t_SINGLEQ1_SINGLEQ2_TRIPLEQ1_TRIPLEQ2_escaped>\\\\(.|\\n))|(?P<t_TRIPLEQ2_simple>[^"\\\\]+)|(?P<t_TRIPLEQ2_q2_but_not_triple>"(?!""))|(?P<t_TRIPLEQ2_end>""")', [None, ('t_SINGLEQ1_SINGLEQ2_TRIPLEQ1_TRIPLEQ2_escaped', 'escaped'), None, ('t_TRIPLEQ2_simple', 'simple'), ('t_TRIPLEQ2_q2_but_not_triple', 'q2_but_not_triple'), ('t_TRIPLEQ2_end', 'end')])], 'TRIPLEQ1': [("(?P<t_SINGLEQ1_SINGLEQ2_TRIPLEQ1_TRIPLEQ2_escaped>\\\\(.|\\n))|(?P<t_TRIPLEQ1_simple>[^'\\\\]+)|(?P<t_TRIPLEQ1_q1_but_not_triple>'(?!''))|(?P<t_TRIPLEQ1_end>''')", [None, ('t_SINGLEQ1_SINGLEQ2_TRIPLEQ1_TRIPLEQ2_escaped', 'escaped'), None, ('t_TRIPLEQ1_simple', 'simple'), ('t_TRIPLEQ1_q1_but_not_triple', 'q1_but_not_triple'), ('t_TRIPLEQ1_end', 'end')])], 'INITIAL': [('(?P<t_NUMBER>((\\d+[jJ]|((\\d+\\.\\d*|\\.\\d+)([eE][-+]?\\d+)?|\\d+[eE][-+]?\\d+)[jJ])|((\\d+\\.\\d*|\\.\\d+)([eE][-+]?\\d+)?|\\d+[eE][-+]?\\d+)|(0[xX][\\da-fA-F]+[lL]?|0[bB][01]+[lL]?|(0[oO][0-7]+)|(0[0-7]*)[lL]?|[1-9]\\d*[lL]?)))|(?P<t_comment>[ ]*\\#[^\\r\\n]*)|(?P<t_WS> [ \\t\\f]+ )|(?P<t_escaped_newline>\\\\\\n)|(?P<t_newline>\\n+)|(?P<t_LPAR>\\()|(?P<t_RPAR>\\))|(?P<t_LBRACE>\\{)|(?P<t_RBRACE>\\})|(?P<t_LSQB>\\[)|(?P<t_RSQB>\\])|(?P<t_start_triple_quoted_q1_string>(([fF]?[uU]?[bB]?[rR]?)|([rR]?[fF]?[uU]?[bB]?))\'\'\')|(?P<t_start_triple_quoted_q2_string>(([fF]?[uU]?[bB]?[rR]?)|([rR]?[fF]?[uU]?[bB]?))""")|(?P<t_start_single_quoted_q1_string>(([fF]?[uU]?[bB]?[rR]?)|([rR]?[fF]?[uU]?[bB]?))\')|(?P<t_start_single_quoted_q2_string>(([fF]?[uU]?[bB]?[rR]?)|([rR]?[fF]?[uU]?[bB]?))")|(?P<t_NAME>[a-zA-Z_]\\w*)|(?P<t_PRAGMA>\\$pragma)|(?P<t_ELLIPSIS>\\.\\.\\.)|(?P<t_DOUBLESTAREQUAL>\\*\\*=)|(?P<t_DOUBLESTAR>\\*\\*)|(?P<t_LEFTSHIFTEQUAL><<=)|(?P<t_RIGHTSHIFTEQUAL>>>=)|(?P<t_VBAREQUAL>\\|=)|(?P<t_STAREQUAL>\\*=)|(?P<t_CIRCUMFLEXEQUAL>\\^=)|(?P<t_DOUBLESLASHEQUAL>//=)|(?P<t_PLUSEQUAL>\\+=)|(?P<t_VBAR>\\|)|(?P<t_LEFTSHIFT><<)|(?P<t_EQEQUAL>==)|(?P<t_PLUS>\\+)|(?P<t_PERCENTEQUAL>%=)|(?P<t_SLASHEQUAL>/=)|(?P<t_RIGHTARROW>=>)|(?P<t_COLONEQUAL>:=)|(?P<t_NOTEQUAL>!=)|(?P<t_STAR>\\*)|(?P<t_GREATEREQUAL>>=)|(?P<t_CIRCUMFLEX>\\^)|(?P<t_DOUBLESLASH>//)|(?P<t_DOT>\\.)|(?P<t_MINUSEQUAL>-=)|(?P<t_DOUBLECOLON>::)|(?P<t_AMPEREQUAL>&=)|(?P<t_RIGHTSHIFT>>>)|(?P<t_LESSEQUAL><=)|(?P<t_EQUAL>=)|(?P<t_AMPER>&)|(?P<t_SLASH>/)|(?P<t_GREATER>>)|(?P<t_LESS><)|(?P<t_COMMA>,)|(?P<t_PERCENT>%)|(?P<t_TILDE>~)|(?P<t_SEMI>;)|(?P<t_MINUS>-)|(?P<t_COLON>:)|(?P<t_AT>@)', [None, ('t_NUMBER', 'NUMBER'), None, None, None, None, None, None, None, None, None, None, None, ('t_comment', 'comment'), ('t_WS', 'WS'), ('t_escaped_newline', 'escaped_newline'), ('t_newline', 'newline'), ('t_LPAR', 'LPAR'), ('t_RPAR', 'RPAR'), ('t_LBRACE', 'LBRACE'), ('t_RBRACE', 'RBRACE'), ('t_LSQB', 'LSQB'), ('t_RSQB', 'RSQB'), ('t_start_triple_quoted_q1_string', 'start_triple_quoted_q1_string'), None, None, None, ('t_start_triple_quoted_q2_string', 'start_triple_quoted_q2_string'), None, None, None, ('t_start_single_quoted_q1_string', 'start_single_quoted_q1_string'), None, None, None, ('t_start_single_quoted_q2_string', 'start_single_quoted_q2_string'), None, None, None, ('t_NAME', 'NAME'), (None, 'PRAGMA'), (None, 'ELLIPSIS'), (None, 'DOUBLESTAREQUAL'), (None, 'DOUBLESTAR'), (None, 'LEFTSHIFTEQUAL'), (None, 'RIGHTSHIFTEQUAL'), (None, 'VBAREQUAL'), (None, 'STAREQUAL'), (None, 'CIRCUMFLEXEQUAL'), (None, 'DOUBLESLASHEQUAL'), (None, 'PLUSEQUAL'), (None, 'VBAR'), (None, 'LEFTSHIFT'), (None, 'EQEQUAL'), (None, 'PLUS'), (None, 'PERCENTEQUAL'), (None, 'SLASHEQUAL'), (None, 'RIGHTARROW'), (None, 'COLONEQUAL'), (None, 'NOTEQUAL'), (None, 'STAR'), (None, 'GREATEREQUAL'), (None, 'CIRCUMFLEX'), (None, 'DOUBLESLASH'), (None, 'DOT'), (None, 'MINUSEQUAL'), (None, 'DOUBLECOLON'), (None, 'AMPEREQUAL'), (None, 'RIGHTSHIFT'), (None, 'LESSEQUAL'), (None, 'EQUAL'), (None, 'AMPER'), (None, 'SLASH'), (None, 'GREATER'), (None, 'LESS'), (None, 'COMMA'), (None, 'PERCENT'), (None, 'TILDE'), (None, 'SEMI'), (None, 'MINUS'), (None, 'COLON'), (None, 'AT')])], 'SINGLEQ2': [('(?P<t_SINGLEQ1_SINGLEQ2_TRIPLEQ1_TRIPLEQ2_escaped>\\\\(.|\\n))|(?P<t_SINGLEQ2_simple>[^"\\\\\\n]+)|(?P<t_SINGLEQ2_end>")', [None, ('t_SINGLEQ1_SINGLEQ2_TRIPLEQ1_TRIPLEQ2_escaped', 'escaped'), None, ('t_SINGLEQ2_simple', 'simple'), ('t_SINGLEQ2_end', 'end')])], 'SINGLEQ1': [("(?P<t_SINGLEQ1_SINGLEQ2_TRIPLEQ1_TRIPLEQ2_escaped>\\\\(.|\\n))|(?P<t_SINGLEQ1_simple>[^'\\\\\\n]+)|(?P<t_SINGLEQ1_end>')", [None, ('t_SINGLEQ1_SINGLEQ2_TRIPLEQ1_TRIPLEQ2_escaped', 'escaped'), None, ('t_SINGLEQ1_simple', 'simple'), ('t_SINGLEQ1_end', 'end')])]}
_lexstateignore = {'TRIPLEQ2': '', 'TRIPLEQ1': '', 'INITIAL': '', 'SINGLEQ2': '', 'SINGLEQ1': ''}
_lexstateerrorf = {'TRIPLEQ2': 't_TRIPLEQ2_error', 'TRIPLEQ1': 't_TRIPLEQ1_error', 'INITIAL': 't_error', 'SINGLEQ2': 't_SINGLEQ2_error', 'SINGLEQ1': 't_SINGLEQ1_error'}
_lexstateeoff = {}



#------------------------------------------------------------------------------
# Copyright (c) 2013, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#------------------------------------------------------------------------------
""" A utility module for dealing with CSS3 font strings.

"""
from past.builtins import basestring

from atom.api import Coerced

from .fontext import Font, FontStyle, FontCaps


#: A mapping from CSS font style keyword to style enum
_STYLES = {
    'normal': FontStyle.Normal,
    'italic': FontStyle.Italic,
    'oblique': FontStyle.Oblique,
}


#: A mapping from CSS font variant keyword to caps enum
_VARIANTS = {
    'normal': FontCaps.MixedCase,
    'small-caps':  FontCaps.SmallCaps,
}


#: A mapping from CSS font weight to integer weight. These values are
#: pulled from the Qt stylesheet implementation of font parsing. Enaml
#: does not support the 'bolder' and 'lighter' keywords.
_WEIGHTS = {
    '100': 12,
    '200': 25,
    '300': 37,
    '400': 50,
    '500': 62,
    '600': 75,
    '700': 87,
    '800': 99,
    '900': 99,
    'normal': 50,
    'bold': 75,
}


#: A mapping from CSS font size keywords to font point sizes. These are
#: based on a standard 12 point font size.
_SIZES = {
    'xx-small': 7,
    'x-small': 8,
    'small': 9,
    'medium': 12,
    'large': 14,
    'x-large': 18,
    'xx-large': 24,
}


#: A mapping from CSS font units to functions which convert to points.
_UNITS = {
    'in': lambda size: int(size * 72.0),
    'cm': lambda size: int(size * 72.0 / 2.54),
    'mm': lambda size: int(size * 72.0 / 254.0),
    'pt': lambda size: int(size),
    'pc': lambda size: int(size * 12.0),
    'px': lambda size: int(size * 0.75),
}


def parse_font(font):
    """ Parse a CSS3 shorthand font string into an Enaml Font object.

    Returns
    -------
    result : Font or None
        A font object representing the parsed font. If the string is
        invalid, None will be returned.

    """
    token = []
    tokens = []
    quotechar = None
    for char in font:
        if quotechar is not None:
            if char == quotechar:
                tokens.append(''.join(token))
                token = []
                quotechar = None
            else:
                token.append(char)
        elif char == '"' or char == "'":
            quotechar = char
        elif char in ' \t':
            if token:
                tokens.append(''.join(token))
                token = []
        else:
            token.append(char)

    # Failed to close quoted string.
    if quotechar is not None:
        return

    if token:
        tokens.append(''.join(token))

    sizes = []
    families = []
    optionals = []
    for token in tokens:
        if token in _STYLES or token in _VARIANTS or token in _WEIGHTS:
            if len(sizes) > 0 or len(families) > 0:
                return None
            optionals.append(token)
        elif token in _SIZES or token[-2:] in _UNITS:
            if len(families) > 0:
                return None
            sizes.append(token)
        else:
            families.append(token)

    if len(optionals) > 3:
        return None
    if len(sizes) != 1:
        return None
    if len(families) != 1:
        return None

    style = None
    variant = None
    weight = None

    for opt in optionals:
        if opt == 'normal':
            continue
        elif opt in _STYLES:
            if style is not None:
                return None
            style = opt
        elif opt in _VARIANTS:
            if variant is not None:
                return None
            variant = opt
        elif opt in _WEIGHTS:
            if weight is not None:
                return None
            weight = opt
        else:
            return None

    size = sizes[0]
    if size in _SIZES:
        size = _SIZES[size]
    else:
        sizenum, units = size[:-2], size[-2:]
        try:
            sizenum = float(sizenum)
        except ValueError:
            return None
        size = _UNITS[units](sizenum)

    family = str(families[0])
    weight = _WEIGHTS[weight] if weight else _WEIGHTS['normal']
    style = _STYLES[style] if style else _STYLES['normal']
    variant = _VARIANTS[variant] if variant else _VARIANTS['normal']

    return Font(family, size, weight, style, variant)


def coerce_font(font):
    """ The coercing function for the FontMember.

    """
    if isinstance(font, basestring):
        return parse_font(font)


class FontMember(Coerced):
    """ An Atom member class which coerces a value to a font.

    A font member can be set to a Font, a string, or None. A string
    font will be parsed into a Font object. If the parsing fails,
    the font will be None.  Font strings must be given in CSS grammar,
    e.g. 'bold 12pt arial', which is order dependant.

    """
    __slots__ = ()

    def __init__(self, default=None, factory=None):
        """ Initialize a FontMember.

        default : Font, string, or None, optional
            The default font to use for the member.

        factory : callable, optional
            An optional callable which takes no arguments and returns
            the default value for the member. If this is provided, it
            will override any value passed as 'default'.

        Notes
        -----
        When providing a default font value, prefer using a Font object
        directly as this object will be shared among all instances of
        the class. Using a font string will result in a new Font object
        being created for each class instance.

        """
        if factory is None:
            factory = lambda: default
        kind = (Font, type(None))
        sup = super(FontMember, self)
        sup.__init__(kind, factory=factory, coercer=coerce_font)



#------------------------------------------------------------------------------
# Copyright (c) 2013-2017, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#------------------------------------------------------------------------------
from atom.api import Int, Typed

from enaml.colors import Color
from enaml.widgets.color_dialog import ProxyColorDialog

from .QtCore import Signal
from .QtGui import QColor
from .QtWidgets import QColorDialog

from .qt_toolkit_dialog import QtToolkitDialog


def color_from_qcolor(q):
    """ Convert a QColor into an Enaml Color.

    Parameters
    ----------
    q : QColor
        The Qt color to convert to Enaml Color.

    Returns
    -------
    result : Color or None
        An Enaml Color or None if the QColor is not valid.

    """
    if not q.isValid():
        return None
    return Color(q.red(), q.green(), q.blue(), q.alpha())


# Guard flags
CURRENT_GUARD = 0x1


class QColorDialogEx(QColorDialog):
    """ A custom QColorDialog which emits a custom finished signal.

    """
    #: A signal emitted at the end of the 'done' method. This works
    #: around the standard QColorDialog behavior which emits the
    #: 'colorSelected' signal *after* the 'finished' signal.
    reallyFinished = Signal(int)

    def done(self, result):
        """ A reimplemented done method.

        This method emits the 'reallyFinished' signal on completion.

        """
        super(QColorDialogEx, self).done(result)
        self.reallyFinished.emit(result)


class QtColorDialog(QtToolkitDialog, ProxyColorDialog):
    """ A Qt implementation of an Enaml ProxyColorDialog.

    """
    #: A reference to the widget created by the proxy.
    widget = Typed(QColorDialogEx)

    #: Cyclic notification guard. This a bitfield of multiple guards.
    _guard = Int(0)

    def create_widget(self):
        """ Create the underlying QColorDialog.

        """
        self.widget = QColorDialogEx(self.parent_widget())

    def init_widget(self):
        """ Initialize the underlying widget.

        """
        super(QtColorDialog, self).init_widget()
        d = self.declaration
        self.set_current_color(d.current_color)
        self.set_show_alpha(d.show_alpha)
        self.set_show_buttons(d.show_buttons)
        widget = self.widget
        widget.currentColorChanged.connect(self.on_current_color_changed)
        widget.colorSelected.connect(self.on_color_selected)
        # use the custom finished signal instead of the superclass'
        widget.finished.disconnect(self.on_finished)
        widget.reallyFinished.connect(self.on_finished)

    #--------------------------------------------------------------------------
    # Utility Methods
    #--------------------------------------------------------------------------
    def get_default_title(self):
        """ Get the default window title for the color dialog.

        """
        return u'Select Color'

    #--------------------------------------------------------------------------
    # Signal Handlers
    #--------------------------------------------------------------------------
    def on_current_color_changed(self, qcolor):
        """ Handle the 'currentColorChanged' signal from the widget.

        """
        d = self.declaration
        if d is not None:
            self._guard |= CURRENT_GUARD
            try:
                d.current_color = color_from_qcolor(qcolor)
            finally:
                self._guard &= ~CURRENT_GUARD

    def on_color_selected(self, qcolor):
        """ Handle the 'colorSelected' signal from the widget.

        """
        d = self.declaration
        if d is not None:
            d.selected_color = color_from_qcolor(qcolor)

    #--------------------------------------------------------------------------
    # ProxyColorDialog API
    #--------------------------------------------------------------------------
    @staticmethod
    def custom_count():
        """ Get the number of available custom colors.

        """
        return QColorDialog.customCount()

    @staticmethod
    def custom_color(index):
        """ Get the custom color for the given index.

        """
        qrgb = QColorDialog.customColor(index)
        return color_from_qcolor(QColor.fromRgba(qrgb))

    @staticmethod
    def set_custom_color(index, color):
        """ Set the custom color for the given index.

        """
        QColorDialog.setCustomColor(index, color.argb)

    def set_current_color(self, color):
        """ Set the current color for the underlying widget.

        """
        if not self._guard & CURRENT_GUARD:
            if color is not None:
                qcolor = QColor.fromRgba(color.argb)
            else:
                qcolor = QColor()
            self.widget.setCurrentColor(qcolor)

    def set_show_alpha(self, show):
        """ Set the show alpha option on the underlying widget.

        """
        widget = self.widget
        opt = widget.options()
        if show:
            opt |= QColorDialog.ShowAlphaChannel
        else:
            opt &= ~QColorDialog.ShowAlphaChannel
        widget.setOptions(opt)

    def set_show_buttons(self, show):
        """ Set the show buttons option on the underlying widget.

        """
        widget = self.widget
        opt = widget.options()
        if show:
            opt &= ~QColorDialog.NoButtons
        else:
            opt |= QColorDialog.NoButtons
        widget.setOptions(opt)



#------------------------------------------------------------------------------
# Copyright (c) 2013-2017, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#------------------------------------------------------------------------------
#: A static mapping of language theme token name to lexer attribute name.
TOKENS = {
    "bash": {
        "backticks": "Backticks",
        "comment": "Comment",
        "default": "Default",
        "double_quoted_string": "DoubleQuotedString",
        "error": "Error",
        "here_document_delimiter": "HereDocumentDelimiter",
        "identifier": "Identifier",
        "keyword": "Keyword",
        "number": "Number",
        "operator": "Operator",
        "parameter_expansion": "ParameterExpansion",
        "scalar": "Scalar",
        "single_quoted_here_document": "SingleQuotedHereDocument",
        "single_quoted_string": "SingleQuotedString"
    },
    "batch": {
        "comment": "Comment",
        "default": "Default",
        "external_command": "ExternalCommand",
        "hide_command_char": "HideCommandChar",
        "keyword": "Keyword",
        "label": "Label",
        "operator": "Operator",
        "variable": "Variable"
    },
    "cmake": {
        "block_foreach": "BlockForeach",
        "block_if": "BlockIf",
        "block_macro": "BlockMacro",
        "block_while": "BlockWhile",
        "comment": "Comment",
        "default": "Default",
        "function": "Function",
        "keyword_set3": "KeywordSet3",
        "label": "Label",
        "number": "Number",
        "string": "String",
        "string_left_quote": "StringLeftQuote",
        "string_right_quote": "StringRightQuote",
        "string_variable": "StringVariable",
        "variable": "Variable"
    },
    "cpp": {
        "comment": "Comment",
        "comment_doc": "CommentDoc",
        "comment_doc_keyword": "CommentDocKeyword",
        "comment_doc_keyword_error": "CommentDocKeywordError",
        "comment_line": "CommentLine",
        "comment_line_doc": "CommentLineDoc",
        "default": "Default",
        "double_quoted_string": "DoubleQuotedString",
        "global_class": "GlobalClass",
        "hash_quoted_string": "HashQuotedString",
        "identifier": "Identifier",
        "inactive_comment": "InactiveComment",
        "inactive_comment_doc": "InactiveCommentDoc",
        "inactive_comment_doc_keyword": "InactiveCommentDocKeyword",
        "inactive_comment_doc_keyword_error": "InactiveCommentDocKeywordError",
        "inactive_comment_line": "InactiveCommentLine",
        "inactive_comment_line_doc": "InactiveCommentLineDoc",
        "inactive_default": "InactiveDefault",
        "inactive_double_quoted_string": "InactiveDoubleQuotedString",
        "inactive_global_class": "InactiveGlobalClass",
        "inactive_hash_quoted_string": "InactiveHashQuotedString",
        "inactive_identifier": "InactiveIdentifier",
        "inactive_keyword": "InactiveKeyword",
        "inactive_keyword_set2": "InactiveKeywordSet2",
        "inactive_number": "InactiveNumber",
        "inactive_operator": "InactiveOperator",
        "inactive_pre_processor": "InactivePreProcessor",
        "inactive_pre_processor_comment": "InactivePreProcessorComment",
        "inactive_raw_string": "InactiveRawString",
        "inactive_regex": "InactiveRegex",
        "inactive_single_quoted_string": "InactiveSingleQuotedString",
        "inactive_triple_quoted_verbatim_string": "InactiveTripleQuotedVerbatimString",
        "inactive_unclosed_string": "InactiveUnclosedString",
        "inactive_uuid": "InactiveUUID",
        "inactive_verbatim_string": "InactiveVerbatimString",
        "keyword": "Keyword",
        "keyword_set2": "KeywordSet2",
        "number": "Number",
        "operator": "Operator",
        "pre_processor": "PreProcessor",
        "pre_processor_comment": "PreProcessorComment",
        "raw_string": "RawString",
        "regex": "Regex",
        "single_quoted_string": "SingleQuotedString",
        "triple_quoted_verbatim_string": "TripleQuotedVerbatimString",
        "unclosed_string": "UnclosedString",
        "uuid": "UUID",
        "verbatim_string": "VerbatimString"
    },
    "csharp": {
        # this uses the "cpp" token set
    },
    "css": {
        "at_rule": "AtRule",
        "attribute": "Attribute",
        "class_selector": "ClassSelector",
        "comment": "Comment",
        "css1_property": "CSS1Property",
        "css2_property": "CSS2Property",
        "css3_property": "CSS3Property",
        "default": "Default",
        "double_quoted_string": "DoubleQuotedString",
        "extended_css_property": "ExtendedCSSProperty",
        "extended_pseudo_class": "ExtendedPseudoClass",
        "extended_pseudo_element": "ExtendedPseudoElement",
        "id_selector": "IDSelector",
        "important": "Important",
        "media_rule": "MediaRule",
        "operator": "Operator",
        "pseudo_class": "PseudoClass",
        "pseudo_element": "PseudoElement",
        "single_quoted_string": "SingleQuotedString",
        "tag": "Tag",
        "unknown_property": "UnknownProperty",
        "unknown_pseudo_class": "UnknownPseudoClass",
        "value": "Value",
        "variable": "Variable"
    },
    "d": {
        "backquote_string": "BackquoteString",
        "character": "Character",
        "comment": "Comment",
        "comment_doc": "CommentDoc",
        "comment_doc_keyword": "CommentDocKeyword",
        "comment_doc_keyword_error": "CommentDocKeywordError",
        "comment_line": "CommentLine",
        "comment_line_doc": "CommentLineDoc",
        "comment_nested": "CommentNested",
        "default": "Default",
        "identifier": "Identifier",
        "keyword": "Keyword",
        "keyword_doc": "KeywordDoc",
        "keyword_secondary": "KeywordSecondary",
        "keyword_set5": "KeywordSet5",
        "keyword_set6": "KeywordSet6",
        "keyword_set7": "KeywordSet7",
        "number": "Number",
        "operator": "Operator",
        "raw_string": "RawString",
        "string": "String",
        "typedefs": "Typedefs",
        "unclosed_string": "UnclosedString"
    },
    "diff": {
        "command": "Command",
        "comment": "Comment",
        "default": "Default",
        "header": "Header",
        "line_added": "LineAdded",
        "line_changed": "LineChanged",
        "line_removed": "LineRemoved",
        "position": "Position"
    },
    "enaml": {
        # this uses the "python" token set
    },
    "fortran": {
        # this uses the "fortran77" token set
    },
    "fortran77": {
        "comment": "Comment",
        "continuation": "Continuation",
        "default": "Default",
        "dotted_operator": "DottedOperator",
        "double_quoted_string": "DoubleQuotedString",
        "extended_function": "ExtendedFunction",
        "identifier": "Identifier",
        "intrinsic_function": "IntrinsicFunction",
        "keyword": "Keyword",
        "label": "Label",
        "number": "Number",
        "operator": "Operator",
        "pre_processor": "PreProcessor",
        "single_quoted_string": "SingleQuotedString",
        "unclosed_string": "UnclosedString"
    },
    "html": {
        "asp_at_start": "ASPAtStart",
        "asp_java_script_comment": "ASPJavaScriptComment",
        "asp_java_script_comment_doc": "ASPJavaScriptCommentDoc",
        "asp_java_script_comment_line": "ASPJavaScriptCommentLine",
        "asp_java_script_default": "ASPJavaScriptDefault",
        "asp_java_script_double_quoted_string": "ASPJavaScriptDoubleQuotedString",
        "asp_java_script_keyword": "ASPJavaScriptKeyword",
        "asp_java_script_number": "ASPJavaScriptNumber",
        "asp_java_script_regex": "ASPJavaScriptRegex",
        "asp_java_script_single_quoted_string": "ASPJavaScriptSingleQuotedString",
        "asp_java_script_start": "ASPJavaScriptStart",
        "asp_java_script_symbol": "ASPJavaScriptSymbol",
        "asp_java_script_unclosed_string": "ASPJavaScriptUnclosedString",
        "asp_java_script_word": "ASPJavaScriptWord",
        "asp_python_class_name": "ASPPythonClassName",
        "asp_python_comment": "ASPPythonComment",
        "asp_python_default": "ASPPythonDefault",
        "asp_python_double_quoted_string": "ASPPythonDoubleQuotedString",
        "asp_python_function_method_name": "ASPPythonFunctionMethodName",
        "asp_python_identifier": "ASPPythonIdentifier",
        "asp_python_keyword": "ASPPythonKeyword",
        "asp_python_number": "ASPPythonNumber",
        "asp_python_operator": "ASPPythonOperator",
        "asp_python_single_quoted_string": "ASPPythonSingleQuotedString",
        "asp_python_start": "ASPPythonStart",
        "asp_python_triple_double_quoted_string": "ASPPythonTripleDoubleQuotedString",
        "asp_python_triple_single_quoted_string": "ASPPythonTripleSingleQuotedString",
        "asp_start": "ASPStart",
        "aspvb_script_comment": "ASPVBScriptComment",
        "aspvb_script_default": "ASPVBScriptDefault",
        "aspvb_script_identifier": "ASPVBScriptIdentifier",
        "aspvb_script_keyword": "ASPVBScriptKeyword",
        "aspvb_script_number": "ASPVBScriptNumber",
        "aspvb_script_start": "ASPVBScriptStart",
        "aspvb_script_string": "ASPVBScriptString",
        "aspvb_script_unclosed_string": "ASPVBScriptUnclosedString",
        "aspxc_comment": "ASPXCComment",
        "attribute": "Attribute",
        "cdata": "CDATA",
        "default": "Default",
        "entity": "Entity",
        "html_comment": "HTMLComment",
        "html_double_quoted_string": "HTMLDoubleQuotedString",
        "html_number": "HTMLNumber",
        "html_single_quoted_string": "HTMLSingleQuotedString",
        "html_value": "HTMLValue",
        "java_script_comment": "JavaScriptComment",
        "java_script_comment_doc": "JavaScriptCommentDoc",
        "java_script_comment_line": "JavaScriptCommentLine",
        "java_script_default": "JavaScriptDefault",
        "java_script_double_quoted_string": "JavaScriptDoubleQuotedString",
        "java_script_keyword": "JavaScriptKeyword",
        "java_script_number": "JavaScriptNumber",
        "java_script_regex": "JavaScriptRegex",
        "java_script_single_quoted_string": "JavaScriptSingleQuotedString",
        "java_script_start": "JavaScriptStart",
        "java_script_symbol": "JavaScriptSymbol",
        "java_script_unclosed_string": "JavaScriptUnclosedString",
        "java_script_word": "JavaScriptWord",
        "other_in_tag": "OtherInTag",
        "php_comment": "PHPComment",
        "php_comment_line": "PHPCommentLine",
        "php_default": "PHPDefault",
        "php_double_quoted_string": "PHPDoubleQuotedString",
        "php_double_quoted_variable": "PHPDoubleQuotedVariable",
        "php_keyword": "PHPKeyword",
        "php_number": "PHPNumber",
        "php_operator": "PHPOperator",
        "php_single_quoted_string": "PHPSingleQuotedString",
        "php_start": "PHPStart",
        "php_variable": "PHPVariable",
        "python_class_name": "PythonClassName",
        "python_comment": "PythonComment",
        "python_default": "PythonDefault",
        "python_double_quoted_string": "PythonDoubleQuotedString",
        "python_function_method_name": "PythonFunctionMethodName",
        "python_identifier": "PythonIdentifier",
        "python_keyword": "PythonKeyword",
        "python_number": "PythonNumber",
        "python_operator": "PythonOperator",
        "python_single_quoted_string": "PythonSingleQuotedString",
        "python_start": "PythonStart",
        "python_triple_double_quoted_string": "PythonTripleDoubleQuotedString",
        "python_triple_single_quoted_string": "PythonTripleSingleQuotedString",
        "script": "Script",
        "sgml_block_default": "SGMLBlockDefault",
        "sgml_command": "SGMLCommand",
        "sgml_comment": "SGMLComment",
        "sgml_default": "SGMLDefault",
        "sgml_double_quoted_string": "SGMLDoubleQuotedString",
        "sgml_entity": "SGMLEntity",
        "sgml_error": "SGMLError",
        "sgml_parameter": "SGMLParameter",
        "sgml_parameter_comment": "SGMLParameterComment",
        "sgml_single_quoted_string": "SGMLSingleQuotedString",
        "sgml_special": "SGMLSpecial",
        "tag": "Tag",
        "unknown_attribute": "UnknownAttribute",
        "unknown_tag": "UnknownTag",
        "vb_script_comment": "VBScriptComment",
        "vb_script_default": "VBScriptDefault",
        "vb_script_identifier": "VBScriptIdentifier",
        "vb_script_keyword": "VBScriptKeyword",
        "vb_script_number": "VBScriptNumber",
        "vb_script_start": "VBScriptStart",
        "vb_script_string": "VBScriptString",
        "vb_script_unclosed_string": "VBScriptUnclosedString",
        "xml_end": "XMLEnd",
        "xml_start": "XMLStart",
        "xml_tag_end": "XMLTagEnd"
    },
    "idl": {
        # this uses the "cpp" token set
    },
    "java": {
        # this uses the "cpp" token set
    },
    "javascript": {
        # this uses the "cpp" token set
    },
    "lua": {
        "basic_functions": "BasicFunctions",
        "character": "Character",
        "comment": "Comment",
        "coroutines_io_system_facilities": "CoroutinesIOSystemFacilities",
        "default": "Default",
        "identifier": "Identifier",
        "keyword": "Keyword",
        "keyword_set5": "KeywordSet5",
        "keyword_set6": "KeywordSet6",
        "keyword_set7": "KeywordSet7",
        "keyword_set8": "KeywordSet8",
        "label": "Label",
        "line_comment": "LineComment",
        "literal_string": "LiteralString",
        "number": "Number",
        "operator": "Operator",
        "preprocessor": "Preprocessor",
        "string": "String",
        "string_table_maths_functions": "StringTableMathsFunctions",
        "unclosed_string": "UnclosedString"
    },
    "makefile": {
        "comment": "Comment",
        "default": "Default",
        "error": "Error",
        "operator": "Operator",
        "preprocessor": "Preprocessor",
        "target": "Target",
        "variable": "Variable"
    },
    "matlab": {
        "command": "Command",
        "comment": "Comment",
        "default": "Default",
        "double_quoted_string": "DoubleQuotedString",
        "identifier": "Identifier",
        "keyword": "Keyword",
        "number": "Number",
        "operator": "Operator",
        "single_quoted_string": "SingleQuotedString"
    },
    "octave": {
        # this uses the "matlab" token set
    },
    "pascal": {
        "asm": "Asm",
        "character": "Character",
        "comment": "Comment",
        "comment_line": "CommentLine",
        "comment_parenthesis": "CommentParenthesis",
        "default": "Default",
        "hex_number": "HexNumber",
        "identifier": "Identifier",
        "keyword": "Keyword",
        "number": "Number",
        "operator": "Operator",
        "pre_processor": "PreProcessor",
        "pre_processor_parenthesis": "PreProcessorParenthesis",
        "single_quoted_string": "SingleQuotedString",
        "unclosed_string": "UnclosedString"
    },
    "perl": {
        "array": "Array",
        "backtick_here_document": "BacktickHereDocument",
        "backtick_here_document_var": "BacktickHereDocumentVar",
        "backticks": "Backticks",
        "backticks_var": "BackticksVar",
        "comment": "Comment",
        "data_section": "DataSection",
        "default": "Default",
        "double_quoted_here_document": "DoubleQuotedHereDocument",
        "double_quoted_here_document_var": "DoubleQuotedHereDocumentVar",
        "double_quoted_string": "DoubleQuotedString",
        "double_quoted_string_var": "DoubleQuotedStringVar",
        "error": "Error",
        "format_body": "FormatBody",
        "format_identifier": "FormatIdentifier",
        "hash": "Hash",
        "here_document_delimiter": "HereDocumentDelimiter",
        "identifier": "Identifier",
        "keyword": "Keyword",
        "number": "Number",
        "operator": "Operator",
        "pod": "POD",
        "pod_verbatim": "PODVerbatim",
        "quoted_string_q": "QuotedStringQ",
        "quoted_string_qq": "QuotedStringQQ",
        "quoted_string_qq_var": "QuotedStringQQVar",
        "quoted_string_qr": "QuotedStringQR",
        "quoted_string_qr_var": "QuotedStringQRVar",
        "quoted_string_qw": "QuotedStringQW",
        "quoted_string_qx": "QuotedStringQX",
        "quoted_string_qx_var": "QuotedStringQXVar",
        "regex": "Regex",
        "regex_var": "RegexVar",
        "scalar": "Scalar",
        "single_quoted_here_document": "SingleQuotedHereDocument",
        "single_quoted_string": "SingleQuotedString",
        "subroutine_prototype": "SubroutinePrototype",
        "substitution": "Substitution",
        "substitution_var": "SubstitutionVar",
        "symbol_table": "SymbolTable",
        "translation": "Translation"
    },
    "postscript": {
        "array_parenthesis": "ArrayParenthesis",
        "bad_string_character": "BadStringCharacter",
        "base85_string": "Base85String",
        "comment": "Comment",
        "default": "Default",
        "dictionary_parenthesis": "DictionaryParenthesis",
        "dsc_comment": "DSCComment",
        "dsc_comment_value": "DSCCommentValue",
        "hex_string": "HexString",
        "immediate_eval_literal": "ImmediateEvalLiteral",
        "keyword": "Keyword",
        "literal": "Literal",
        "name": "Name",
        "number": "Number",
        "procedure_parenthesis": "ProcedureParenthesis",
        "text": "Text"
    },
    "pov": {
        "bad_directive": "BadDirective",
        "comment": "Comment",
        "comment_line": "CommentLine",
        "default": "Default",
        "directive": "Directive",
        "identifier": "Identifier",
        "keyword_set6": "KeywordSet6",
        "keyword_set7": "KeywordSet7",
        "keyword_set8": "KeywordSet8",
        "number": "Number",
        "objects_csg_appearance": "ObjectsCSGAppearance",
        "operator": "Operator",
        "predefined_functions": "PredefinedFunctions",
        "predefined_identifiers": "PredefinedIdentifiers",
        "string": "String",
        "types_modifiers_items": "TypesModifiersItems",
        "unclosed_string": "UnclosedString"
    },
    "properties": {
        "assignment": "Assignment",
        "comment": "Comment",
        "default": "Default",
        "default_value": "DefaultValue",
        "key": "Key",
        "section": "Section"
    },
    "python": {
        "class_name": "ClassName",
        "comment": "Comment",
        "comment_block": "CommentBlock",
        "decorator": "Decorator",
        "default": "Default",
        "double_quoted_string": "DoubleQuotedString",
        "function_method_name": "FunctionMethodName",
        "highlighted_identifier": "HighlightedIdentifier",
        "identifier": "Identifier",
        "keyword": "Keyword",
        "number": "Number",
        "operator": "Operator",
        "single_quoted_string": "SingleQuotedString",
        "triple_double_quoted_string": "TripleDoubleQuotedString",
        "triple_single_quoted_string": "TripleSingleQuotedString",
        "unclosed_string": "UnclosedString"
    },
    "ruby": {
        "backticks": "Backticks",
        "class_name": "ClassName",
        "class_variable": "ClassVariable",
        "comment": "Comment",
        "data_section": "DataSection",
        "default": "Default",
        "demoted_keyword": "DemotedKeyword",
        "double_quoted_string": "DoubleQuotedString",
        "error": "Error",
        "function_method_name": "FunctionMethodName",
        "global": "Global",
        "here_document": "HereDocument",
        "here_document_delimiter": "HereDocumentDelimiter",
        "identifier": "Identifier",
        "instance_variable": "InstanceVariable",
        "keyword": "Keyword",
        "module_name": "ModuleName",
        "number": "Number",
        "operator": "Operator",
        "percent_string_q": "PercentStringQ",
        "percent_stringq": "PercentStringq",
        "percent_stringr": "PercentStringr",
        "percent_stringw": "PercentStringw",
        "percent_stringx": "PercentStringx",
        "pod": "POD",
        "regex": "Regex",
        "single_quoted_string": "SingleQuotedString",
        "stderr": "Stderr",
        "stdin": "Stdin",
        "stdout": "Stdout",
        "symbol": "Symbol"
    },
    "spice": {
        "command": "Command",
        "comment": "Comment",
        "default": "Default",
        "delimiter": "Delimiter",
        "function": "Function",
        "identifier": "Identifier",
        "number": "Number",
        "parameter": "Parameter",
        "value": "Value"
    },
    "sql": {
        "comment": "Comment",
        "comment_doc": "CommentDoc",
        "comment_doc_keyword": "CommentDocKeyword",
        "comment_doc_keyword_error": "CommentDocKeywordError",
        "comment_line": "CommentLine",
        "comment_line_hash": "CommentLineHash",
        "default": "Default",
        "double_quoted_string": "DoubleQuotedString",
        "identifier": "Identifier",
        "keyword": "Keyword",
        "keyword_set5": "KeywordSet5",
        "keyword_set6": "KeywordSet6",
        "keyword_set7": "KeywordSet7",
        "keyword_set8": "KeywordSet8",
        "number": "Number",
        "operator": "Operator",
        "plus_comment": "PlusComment",
        "plus_keyword": "PlusKeyword",
        "plus_prompt": "PlusPrompt",
        "quoted_identifier": "QuotedIdentifier",
        "single_quoted_string": "SingleQuotedString"
    },
    "tcl": {
        "comment": "Comment",
        "comment_block": "CommentBlock",
        "comment_box": "CommentBox",
        "comment_line": "CommentLine",
        "default": "Default",
        "expand_keyword": "ExpandKeyword",
        "identifier": "Identifier",
        "itcl_keyword": "ITCLKeyword",
        "keyword_set6": "KeywordSet6",
        "keyword_set7": "KeywordSet7",
        "keyword_set8": "KeywordSet8",
        "keyword_set9": "KeywordSet9",
        "modifier": "Modifier",
        "number": "Number",
        "operator": "Operator",
        "quoted_keyword": "QuotedKeyword",
        "quoted_string": "QuotedString",
        "substitution": "Substitution",
        "substitution_brace": "SubstitutionBrace",
        "tcl_keyword": "TCLKeyword",
        "tk_command": "TkCommand",
        "tk_keyword": "TkKeyword"
    },
    "tex": {
        "command": "Command",
        "default": "Default",
        "group": "Group",
        "special": "Special",
        "symbol": "Symbol",
        "text": "Text"
    },
    "verilog": {
        "comment": "Comment",
        "comment_bang": "CommentBang",
        "comment_line": "CommentLine",
        "default": "Default",
        "identifier": "Identifier",
        "keyword": "Keyword",
        "keyword_set2": "KeywordSet2",
        "number": "Number",
        "operator": "Operator",
        "preprocessor": "Preprocessor",
        "string": "String",
        "system_task": "SystemTask",
        "unclosed_string": "UnclosedString",
        "user_keyword_set": "UserKeywordSet"
    },
    "vhdl": {
        "attribute": "Attribute",
        "comment": "Comment",
        "comment_line": "CommentLine",
        "default": "Default",
        "identifier": "Identifier",
        "keyword": "Keyword",
        "keyword_set7": "KeywordSet7",
        "number": "Number",
        "operator": "Operator",
        "standard_function": "StandardFunction",
        "standard_operator": "StandardOperator",
        "standard_package": "StandardPackage",
        "standard_type": "StandardType",
        "string": "String",
        "unclosed_string": "UnclosedString"
    },
    "xml": {
        # this uses the "html" token set
    },
    "yaml": {
        "comment": "Comment",
        "default": "Default",
        "document_delimiter": "DocumentDelimiter",
        "identifier": "Identifier",
        "keyword": "Keyword",
        "number": "Number",
        "operator": "Operator",
        "reference": "Reference",
        "syntax_error_marker": "SyntaxErrorMarker",
        "text_block_marker": "TextBlockMarker"
    }
}


TOKENS["csharp"] = TOKENS["cpp"]
TOKENS["enaml"] = TOKENS["python"]
TOKENS["fortran"] = TOKENS["fortran77"]
TOKENS["idl"] = TOKENS["cpp"]
TOKENS["java"] = TOKENS["cpp"]
TOKENS["javascript"] = TOKENS["cpp"]
TOKENS["octave"] = TOKENS["matlab"]
TOKENS["xml"] = TOKENS["html"]
